# Copyright 2020-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
.. contents:: :local:
.. currentmodule:: quantumcompiler

Gate Modules
############

Larger unitary computational modules that
can be broken up into standard gates.

Danger: Gate's may have many qubits, so explicitly creating
the gate tensor may consume huge amounts of memory. Beware.

.. autoclass:: IdentityGate
    :members:

.. autoclass:: PauliGate
    :members:

.. autoclass:: MultiSwapGate
    :members:

.. autoclass:: ReversalGate
    :members:

.. autoclass:: CircularShiftGate
    :members:

.. autoclass:: ControlGate
    :members:

.. autoclass:: QFTGate
    :members:

.. autoclass:: InvQFTGate
    :members:
"""

from typing import Iterable, Iterator, List, Mapping, Union

import networkx as nx
import numpy as np
from networkx.algorithms.approximation.steinertree import steiner_tree
from sympy.combinatorics import Permutation

from . import tensors, utils, var
from .circuits import Circuit
from .gates import unitary_from_hamiltonian
from .ops import Gate, Operation
from .paulialgebra import Pauli, pauli_commuting_sets, sX, sY, sZ
from .qubits import Qubits
from .states import Density, State
from .stdgates import CZ, CNot, CZPow
from .stdgates import H as _H  # NB: Workaround for name conflict with Gate.H
from .stdgates import I, Swap, X, XPow, Y, YPow, Z, ZPow
from .tensors import QubitTensor
from .var import Variable

__all__ = (
    "IdentityGate",
    "PauliGate",
    "MultiSwapGate",
    "ReversalGate",
    "CircularShiftGate",
    "ControlGate",
    "QFTGate",
    "InvQFTGate",
)


class IdentityGate(Gate):
    r"""
    The multi-qubit identity gate.
    """
    cv_interchangeable = True
    cv_hermitian = True
    cv_tensor_structure = "identity"

    @property
    def hamiltonian(self) -> Pauli:
        return Pauli.zero()

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return tensors.asqutensor(np.eye(2 ** self.qubit_nb))

    def __pow__(self, t: Variable) -> "IdentityGate":
        return self

    @property
    def H(self) -> "IdentityGate":
        return self

    def decompose(self) -> Iterator[I]:  # noqa: E741
        for q in self.qubits:
            yield I(q)


# end class IdentityGate


# TESTME : axes
# DOCME: axes
# TODO: diagrams
# TODO: Decompose
# ⊖ ⊕ ⊘ ⊗ ● ○
class ControlGate(Gate):
    """A controlled unitary gate. Given C control qubits and a
    gate acting on K qubits, return a gate with C+K qubits
    """

    def __init__(self, control_qubits: Qubits, gate: Gate, axes: str = None) -> None:
        control_qubits = tuple(control_qubits)
        qubits = tuple(control_qubits) + tuple(gate.qubits)
        if len(set(qubits)) != len(qubits):
            raise ValueError("Control and gate qubits overlap")

        if axes is None:
            axes = "Z" * len(control_qubits)
        assert len(axes) == len(control_qubits)

        super().__init__(qubits)
        self.control_qubits = qubits
        self.gate = gate
        self.axes = axes

    @property
    def hamiltonian(self) -> Pauli:
        ctrlham = {
            "X": (1 - sX(0)) / 2,
            "x": sX(0) / 2,
            "Y": (1 - sY(0)) / 2,
            "y": sY(0) / 2,
            "Z": (1 - sZ(0)) / 2,
            "z": sZ(0) / 2,
        }

        ham = self.gate.hamiltonian
        for q, axis in zip(self.control_qubits, self.axes):
            ham *= ctrlham[axis].on(q)

        return ham

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return unitary_from_hamiltonian(self.hamiltonian, self.qubits).tensor


# end class ControlGate


class MultiSwapGate(Gate):
    """A permutation of qubits. A generalized multi-qubit Swap."""

    cv_tensor_structure = "permutation"
    cv_hermitian = True

    def __init__(self, qubits_in: Qubits, qubits_out: Qubits) -> None:
        if set(qubits_in) != set(qubits_out):
            raise ValueError("Incompatible sets of qubits")

        self.qubits_out = tuple(qubits_out)
        self.qubits_in = tuple(qubits_in)
        super().__init__(qubits=qubits_in)

    @classmethod
    def from_gates(cls, gates: Iterable[Operation]) -> "MultiSwapGate":
        """Create a qubit permutation from a circuit of swap gates"""
        qubits_in = Circuit(gates).qubits
        N = len(qubits_in)

        circ: List[Gate] = []
        for gate in gates:
            if isinstance(gate, Swap):
                circ.append(gate)
            elif isinstance(gate, MultiSwapGate):
                circ.extend(gate.decompose())
            elif isinstance(gate, I) or isinstance(gate, IdentityGate):
                continue
            else:
                raise ValueError("Swap gate must be built from swap gates")

        perm = list(range(N))
        for elem in circ:
            q0, q1 = elem.qubits
            i0 = qubits_in.index(q0)
            i1 = qubits_in.index(q1)
            perm[i1], perm[i0] = perm[i0], perm[i1]

        qubits_out = [qubits_in[p] for p in perm]
        return cls(qubits_in, qubits_out)

    @property
    def H(self) -> "MultiSwapGate":
        return MultiSwapGate(self.qubits_out, self.qubits_in)

    def run(self, ket: State) -> State:
        qubits = ket.qubits
        N = ket.qubit_nb

        perm = list(range(N))
        for q0, q1 in zip(self.qubits_in, self.qubits_out):
            perm[qubits.index(q0)] = qubits.index(q1)

        tensor = tensors.permute(ket.tensor, perm)

        return State(tensor, qubits, ket.memory)

    def evolve(self, rho: Density) -> Density:
        qubits = rho.qubits
        N = rho.qubit_nb

        perm = list(range(N))
        for q0, q1 in zip(self.qubits_in, self.qubits_out):
            perm[qubits.index(q0)] = qubits.index(q1)
        perm.extend([idx + N for idx in perm])

        tensor = tensors.permute(rho.tensor, perm)

        return Density(tensor, qubits, rho.memory)

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        N = self.qubit_nb
        qubits = self.qubits

        perm = list(range(2 * N))
        for q0, q1 in zip(qubits, self.qubits_out):
            perm[qubits.index(q0)] = qubits.index(q1)

        U = np.eye(2 ** N)
        U = np.reshape(U, [2] * 2 * N)
        U = np.transpose(U, perm)
        return tensors.asqutensor(U)

    def decompose(self) -> Iterator[Swap]:
        """
        Returns a swap network for this permutation, assuming all-to-all
        connectivity.
        """
        qubits = self.qubits

        perm = [self.qubits.index(q) for q in self.qubits_out]
        for idx0, idx1 in Permutation(perm).transpositions():
            yield Swap(qubits[idx0], qubits[idx1])


# end class MultiSwapGate


class ReversalGate(MultiSwapGate):
    """A qubit permutation that reverses the order of qubits"""

    def __init__(self, qubits: Qubits) -> None:
        super().__init__(qubits, tuple(reversed(qubits)))

    def decompose(self) -> Iterator[Swap]:
        qubits = self.qubits
        for idx in range(self.qubit_nb // 2):
            yield Swap(qubits[idx], qubits[-idx - 1])


# end class ReversalGate


# DOCME Makes a circular buffer
class CircularShiftGate(MultiSwapGate):
    def __init__(self, qubits: Qubits, shift: int = 1) -> None:
        qubits_in = tuple(qubits)
        nshift = shift % len(qubits)
        qubits_out = qubits_in[nshift:] + qubits_in[:nshift]

        super().__init__(qubits_in, qubits_out)
        self.shift = shift


# end class CircularShiftGate


class QFTGate(Gate):
    """The Quantum Fourier Transform circuit.

    For 3-qubits
    ::
        0: ───H───Z^1/2───Z^1/4───────────────────x───
                  │       │                       │
        1: ───────●───────┼───────H───Z^1/2───────┼───
                          │           │           │
        2: ───────────────●───────────●───────H───x───
    """

    def __init__(self, qubits: Qubits) -> None:
        super().__init__(qubits=qubits)

    @property
    def H(self) -> "InvQFTGate":
        return InvQFTGate(self.qubits)

    def decompose(self) -> Iterator[Union[_H, CZPow, Swap]]:
        qubits = self.qubits
        N = len(qubits)
        for n0 in range(N):
            q0 = qubits[n0]
            yield _H(q0)
            for n1 in range(n0 + 1, N):
                q1 = qubits[n1]
                yield CZ(q1, q0) ** (1 / 2 ** (n1 - n0))
        yield from ReversalGate(qubits).decompose()

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return Circuit(self.decompose()).asgate().on(*self.qubits).tensor


# end class QFTGate


class InvQFTGate(Gate):
    """The inverse Quantum Fourier Transform"""

    def __init__(self, qubits: Qubits) -> None:
        super().__init__(qubits=qubits)

    @property
    def H(self) -> "QFTGate":
        return QFTGate(self.qubits)

    def decompose(self) -> Iterator[Union[_H, CZPow, Swap]]:
        gates = list(QFTGate(self.qubits).decompose())
        yield from (gate.H for gate in gates[::-1])

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return Circuit(self.decompose()).asgate().on(*self.qubits).tensor


# end class InvQFTGate


class PauliGate(Gate):
    """
    A Gate corresponding to the exponential of the Pauli algebra element,
    i.e. exp[-1.0j * alpha * element]
    """

    # Kudos: GEC (2019).

    def __init__(self, element: Pauli, alpha: float) -> None:

        super().__init__(qubits=element.qubits)
        self.element = element
        self.alpha = alpha

    def __str__(self) -> str:
        return f"PauliGate({self.element}, {self.alpha}) {self.qubits}"

    @property
    def H(self) -> "PauliGate":
        return self ** -1

    def __pow__(self, t: Variable) -> "PauliGate":
        return PauliGate(self.element, self.alpha * t)

    @property
    def hamiltonian(self) -> "Pauli":
        return self.alpha * self.element

    def resolve(self, subs: Mapping[str, float]) -> "PauliGate":
        if var.is_symbolic(self.alpha):
            alpha = var.asfloat(self.alpha, subs)
            return PauliGate(self.element, alpha)
        return self

    # TODO: Move main logic to pauli algebra?
    def decompose(
        self, topology: nx.Graph = None
    ) -> Iterator[Union[CNot, XPow, YPow, ZPow]]:
        """
        Returns a Circuit corresponding to the exponential of
        the Pauli algebra element object, i.e. exp[-1.0j * alpha * element]

        If a qubit topology is provided then the returned circuit will
        respect the qubit connectivity, adding swaps as necessary.
        """
        # Kudos: Adapted from pyquil. The topological network is novel.

        circ = Circuit()
        element = self.element
        alpha = self.alpha

        if element.is_identity() or element.is_zero():
            return circ  # pragma: no cover  # TESTME

        # Check that all terms commute
        groups = pauli_commuting_sets(element)
        if len(groups) != 1:
            raise ValueError("Pauli terms do not all commute")

        for qbs, ops, coeff in element:
            if not np.isclose(complex(coeff).imag, 0.0):
                raise ValueError("Pauli term coefficients must be real")
            theta = complex(coeff).real * alpha

            if len(ops) == 0:
                continue

            # TODO: 1-qubit terms special case

            active_qubits = set()
            change_to_z_basis = Circuit()
            for qubit, op in zip(qbs, ops):
                active_qubits.add(qubit)
                if op == "X":
                    change_to_z_basis += Y(qubit) ** -0.5
                elif op == "Y":
                    change_to_z_basis += X(qubit) ** 0.5

            if topology is not None:
                if not nx.is_directed(topology) or not nx.is_arborescence(topology):
                    # An 'arborescence' is a directed tree
                    active_topology = steiner_tree(topology, active_qubits)
                    center = nx.center(active_topology)[0]
                    active_topology = nx.dfs_tree(active_topology, center)
                else:
                    active_topology = topology
            else:
                active_topology = nx.DiGraph()
                nx.add_path(active_topology, reversed(list(active_qubits)))

            cnot_seq = Circuit()
            order = list(reversed(list(nx.topological_sort(active_topology))))
            for q0 in order[:-1]:
                q1 = list(active_topology.pred[q0])[0]
                if q1 not in active_qubits:
                    cnot_seq += Swap(q0, q1)
                    active_qubits.add(q1)
                else:
                    cnot_seq += CNot(q0, q1)

            circ += change_to_z_basis
            circ += cnot_seq
            circ += Z(order[-1]) ** (2 * theta / np.pi)
            circ += cnot_seq.H
            circ += change_to_z_basis.H
        # end term loop

        yield from circ  # type: ignore

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return Circuit(self.decompose()).asgate().on(*self.qubits).tensor


# end class PauliGate


# Fin



# Copyright 2020-, Gavin E. Crooks and contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


import random

import numpy as np
import pytest

import quantumflow as qf
from quantumflow.config import ATOL

from .config_test import REPS


def test_fubini_study_angle() -> None:

    for _ in range(REPS):
        theta = random.uniform(-np.pi, +np.pi)

        ang = qf.fubini_study_angle(qf.I(0).tensor, qf.Rx(theta, 0).su().tensor)
        assert np.isclose(2 * ang / abs(theta), 1.0)

        ang = qf.fubini_study_angle(qf.I(0).tensor, qf.Ry(theta, 0).tensor)
        assert np.isclose(2 * ang / abs(theta), 1.0)

        ang = qf.fubini_study_angle(qf.I(0).tensor, qf.Rz(theta, 0).tensor)
        assert np.isclose(2 * ang / abs(theta), 1.0)

        ang = qf.fubini_study_angle(qf.Swap(0, 1).tensor, qf.PSwap(theta, 0, 1).tensor)

        assert np.isclose(2 * ang / abs(theta), 1.0)

        ang = qf.fubini_study_angle(qf.I(0).tensor, qf.PhaseShift(theta, 0).tensor)
        assert np.isclose(2 * ang / abs(theta), 1.0)

        assert qf.fubini_study_close(qf.Rz(theta, 0).tensor, qf.Rz(theta, 0).tensor)

    for n in range(1, 6):
        eye = qf.IdentityGate(list(range(n)))
        assert np.isclose(qf.fubini_study_angle(eye.tensor, eye.tensor), 0.0)

    with pytest.raises(ValueError):
        qf.fubini_study_angle(qf.RandomGate([1]).tensor, qf.RandomGate([0, 1]).tensor)


def test_fubini_study_angle_states() -> None:
    # The state angle is half angle in Bloch sphere
    angle1 = 0.1324
    ket1 = qf.zero_state(1)
    ket2 = qf.Rx(angle1, 0).run(ket1)
    angle2 = qf.fubini_study_angle(ket1.tensor, ket2.tensor)
    assert np.isclose(angle1, angle2 * 2)


def test_state_angle() -> None:
    ket0 = qf.random_state(1)
    ket1 = qf.random_state(1)
    qf.state_angle(ket0, ket1)

    assert not qf.states_close(ket0, ket1)
    assert qf.states_close(ket0, ket0)


def test_density_angle() -> None:
    rho0 = qf.random_density(1)
    rho1 = qf.random_density(1)
    qf.density_angle(rho0, rho1)

    assert not qf.densities_close(rho0, rho1)
    assert qf.densities_close(rho0, rho0)


def test_gate_angle() -> None:
    gate0 = qf.RandomGate([1])
    gate1 = qf.RandomGate([1])
    qf.gate_angle(gate0, gate1)

    assert not qf.gates_close(gate0, gate1)
    assert qf.gates_close(gate0, gate0)


def test_gates_commute() -> None:
    assert qf.gates_commute(qf.X(0), qf.X(0))
    assert not qf.gates_commute(qf.X(0), qf.T(0))
    assert qf.gates_commute(qf.X(0), qf.T(1))
    assert qf.gates_commute(qf.S(0), qf.T(0))
    assert qf.gates_commute(qf.S(0), qf.T(0))
    assert qf.gates_commute(qf.XX(0.1, 0, 1), qf.X(0))
    assert not qf.gates_commute(qf.ZZ(0.1, 0, 1), qf.X(0))


def test_channel_angle() -> None:
    chan0 = qf.X(0).aschannel()
    chan1 = qf.Y(0).aschannel()
    qf.channel_angle(chan0, chan1)

    assert not qf.channels_close(chan0, chan1)
    assert qf.channels_close(chan0, chan0)


def test_fidelity() -> None:
    rho0 = qf.random_density(4)
    rho1 = qf.random_density(4)

    fid = qf.fidelity(rho0, rho1)
    assert 0.0 <= fid <= 1.0

    rho2 = qf.random_density([3, 2, 1, 0])
    fid = qf.fidelity(rho0, rho2)
    assert 0.0 <= fid <= 1.0

    fid = qf.fidelity(rho0, rho0)
    assert np.isclose(fid, 1.0)

    ket0 = qf.random_state(3)
    ket1 = qf.random_state(3)
    fid0 = qf.state_fidelity(ket0, ket1)

    rho0 = ket0.asdensity()
    rho1 = ket1.asdensity()
    fid1 = qf.fidelity(rho0, rho1)

    assert np.isclose(fid1, fid0)

    fid2 = np.cos(qf.fubini_study_angle(ket0.tensor, ket1.tensor)) ** 2
    assert np.isclose(fid2, fid0)


def test_purity() -> None:
    density = qf.ghz_state(4).asdensity()
    assert np.isclose(qf.purity(density), 1.0)

    for _ in range(10):
        density = qf.random_density(4)
        purity = np.real(qf.purity(density))
        assert purity < 1.0
        assert purity >= 0.0

    rho = qf.Density(np.diag([0.9, 0.1]))
    assert np.isclose(qf.purity(rho), 0.82)  # Kudos: Josh Combs


def test_bures_distance() -> None:
    rho = qf.random_density(4)
    assert np.isclose(qf.bures_distance(rho, rho), 0.0, atol=ATOL * 10)

    rho1 = qf.random_density(4)
    qf.bures_distance(rho, rho1)

    # TODO: Check distance of known special case


def test_bures_angle() -> None:
    rho = qf.random_density(4)
    assert np.isclose(qf.bures_angle(rho, rho), 0.0, atol=ATOL * 10)

    rho1 = qf.random_density(4)
    qf.bures_angle(rho, rho1)

    ket0 = qf.random_state(4)
    ket1 = qf.random_state(4)
    rho0 = ket0.asdensity()
    rho1 = ket1.asdensity()

    ang0 = qf.fubini_study_angle(ket0.tensor, ket1.tensor)
    ang1 = qf.bures_angle(rho0, rho1)

    assert np.isclose(ang0, ang1)


def test_entropy() -> None:
    N = 4
    rho0 = qf.mixed_density(N)
    ent = qf.entropy(rho0, base=2)
    assert np.isclose(ent, N)

    # Entropy invariant to unitary evolution
    chan = qf.RandomGate(range(N)).aschannel()
    rho1 = chan.evolve(rho0)
    ent = qf.entropy(rho1, base=2)
    assert np.isclose(ent, N)


def test_mutual_info() -> None:
    rho0 = qf.mixed_density(4)
    info0 = qf.mutual_info(rho0, qubits0=[0, 1], qubits1=[2, 3])

    # Information invariant to local unitary evolution
    chan = qf.RandomGate(range(2)).aschannel()
    rho1 = chan.evolve(rho0)
    info1 = qf.mutual_info(rho1, qubits0=[0, 1], qubits1=[2, 3])

    assert np.isclose(info0, info1)

    info2 = qf.mutual_info(rho1, qubits0=[0, 1])
    assert np.isclose(info0, info2)


def test_trace_distance() -> None:
    rho = qf.random_density(4)
    assert np.isclose(qf.trace_distance(rho, rho), 0.0)

    rho1 = qf.random_density(4)
    qf.trace_distance(rho, rho1)

    # TODO: Check distance of known special case


def test_average_gate_fidelity() -> None:
    kraus = qf.Damping(0.9, q0=0)
    qf.average_gate_fidelity(kraus)

    qf.average_gate_fidelity(kraus, qf.X(0))
    # TODO: Test actually get correct answer!


def test_circuits_close() -> None:
    circ0 = qf.Circuit([qf.H(0)])
    circ1 = qf.Circuit([qf.H(2)])
    assert not qf.circuits_close(circ0, circ1)

    circ2 = qf.Circuit([qf.X(0)])
    assert not qf.circuits_close(circ0, circ2)

    circ3 = qf.Circuit([qf.H(0)])
    assert qf.circuits_close(circ0, circ3)


def test_gates_phase_close() -> None:
    gate0 = qf.ZPow(0.5, q0=0)
    gate1 = qf.Circuit(qf.translate_tz_to_rz(gate0)).asgate()  # type: ignore
    assert qf.gates_close(gate0, gate1)
    assert not qf.gates_phase_close(gate0, gate1)
    assert qf.gates_phase_close(gate0, gate0)

    gate2 = qf.XPow(0.5, q0=0)
    assert not qf.gates_phase_close(gate0, gate2)


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


# Design Note:
#
# We import the entire public interface to the top level so that users
# don't need to know the internal package layout.
# e.g. `from quantumflow import something`
# or `import quantumflow as qf; qf.something`
#
# We don't include `.utils` since those are internal utility routines.
#
# Submodules for interfacing with external resources (e.g. .xcirq, .xforest, etc.)
# are not imported at top level, since they have external dependencies that
# often break, and need not be installed by default.
#
# We use a star import here (which necessitates suppressing lint messages)
# so that the public API can be specified in the individual submodules.
# Each of these submodules should define __all__.


from .channels import *  # noqa: F401, F403
from .circuits import *  # noqa: F401, F403
from .config import *  # noqa: F401, F403
from .dagcircuit import *  # noqa: F401, F403
from .decompositions import *  # noqa: F401, F403
from .gates import *  # noqa: F401, F403
from .gradients import *  # noqa: F401, F403
from .info import *  # noqa: F401, F403
from .modules import *  # noqa: F401, F403
from .ops import *  # noqa: F401, F403
from .paulialgebra import *  # noqa: F401, F403
from .qubits import *  # noqa: F401, F403
from .states import *  # noqa: F401, F403
from .stdgates import *  # noqa: F401, F403
from .stdops import *  # noqa: F401, F403
from .tensors import *  # noqa: F401, F403
from .transform import *  # noqa: F401, F403
from .translate import *  # noqa: F401, F403
from .var import *  # noqa: F401, F403
from .visualization import *  # noqa: F401, F403

# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
=========
Gradients
=========

QuantumFlow: Gradients of parameterized gates, and gradient descent optimizers

.. contents:: :local:
.. currentmodule:: quantumflow


.. autofunction::   expectation_gradients
.. autofunction::   state_fidelity_gradients
.. autofunction::   state_angle_gradients
.. autofunction::   parameter_shift_circuits

"""

from typing import Callable, Sequence, Tuple

import numpy as np
from numpy import pi

from . import tensors
from .circuits import Circuit
from .gates import join_gates
from .info import state_fidelity
from .ops import Gate, Operation
from .states import State
from .stdgates import XX, YY, ZZ, Rx, Ry, Rz, X, XPow, Y, YPow, Z, ZPow

__all__ = (
    "GRADIENT_GATESET",
    "expectation_gradients",
    "state_fidelity_gradients",
    "state_angle_gradients",
    "parameter_shift_circuits",
    # 'SGD',
    # "Adam",
)


GRADIENT_GATESET = frozenset(
    ["Rx", "Ry", "Rz", "XPow", "YPow", "ZPow", "XX", "YY", "ZZ"]
)
"""
The names of all parameterized gates for which we know how to take gradients.
"""

_UNDIFFERENTIABLE_GATE_MSG = "Undifferentiable gate"

# TODO: Move this data into gate classes?
shift_constant = {
    Rx: 1 / 2,
    Ry: 1 / 2,
    Rz: 1 / 2,
    XPow: pi / 2,
    YPow: pi / 2,
    ZPow: pi / 2,
    XX: pi / 2,
    YY: pi / 2,
    ZZ: pi / 2,
}


gate_generator = {
    Rx: X(0),
    Ry: Y(0),
    Rz: Z(0),
    XPow: X(0),
    YPow: Y(0),
    ZPow: Z(0),
    XX: join_gates(X(0), X(1)),
    YY: join_gates(Y(0), Y(1)),
    ZZ: join_gates(Z(0), Z(1)),
}


def expectation_gradients(
    ket0: State,
    circ: Circuit,
    hermitian: Operation,
    dfunc: Callable[[float], float] = None,
) -> Sequence[float]:
    """
    Calculate the gradients of a function of expectation for a
    parameterized quantum circuit, using the middle-out algorithm.

    Args:
        ket0: An initial state.
        circ: A circuit that acts on the initial state.
        hermitian: A Hermitian Operation for which the expectation is evaluated
        dfunc: Derivative of func. Defaults to identity.

    Returns:
        The gradient with respect to the circuits parameters.
    """
    #  res = func(<ket0|circ^H [hermitian] circ |ket0>)
    grads = []
    forward = ket0
    back = circ.run(ket0)
    print(">>>", back.qubits, hermitian.qubits)
    back = hermitian.run(back)
    back = circ.H.run(back)

    expectation = tensors.inner(forward.tensor, back.tensor)

    for elem in circ:
        assert isinstance(elem, Gate)
        back = elem.run(back)
        forward = elem.run(forward)

        if len(list(elem.params)) == 0:  # Skip gates with no parameters
            continue

        gate_type = type(elem)
        if gate_type not in shift_constant:
            raise ValueError(_UNDIFFERENTIABLE_GATE_MSG)
        r = shift_constant[gate_type]
        gen = gate_generator[gate_type].on(*elem.qubits)

        f0 = gen.run(forward)
        g = -2 * r * np.imag(tensors.inner(f0.tensor, back.tensor))

        if dfunc is not None:
            g = g * dfunc(expectation)

        grads.append(g)

    return grads


def state_fidelity_gradients(
    ket0: State, ket1: State, circ: Circuit
) -> Sequence[float]:
    """
    Calculate the gradients of state fidelity for a parameterized quantum
    circuit, using the middle-out algorithm.

    Args:
        ket0: An initial state.
        ket1: A target state. We calculate the fidelity between this state and
        the resultant of the circuit.
        circ: A circuit that acts on ket0.
    Returns:
        The gradients of state fidelity with respect to the circuits
        parameters.
    """
    grads = []

    forward = ket0
    back = circ.H.run(ket1)
    ol = tensors.inner(forward.tensor, back.tensor)

    for elem in circ:
        assert isinstance(elem, Gate)
        back = elem.run(back)
        forward = elem.run(forward)

        if len(list(elem.params)) == 0:  # Skip gates with no parameters
            continue

        gate_type = type(elem)
        if gate_type not in shift_constant:
            raise ValueError(_UNDIFFERENTIABLE_GATE_MSG + ": " + str(elem))
        r = shift_constant[gate_type]
        gen = gate_generator[gate_type].on(*elem.qubits)

        f0 = gen.run(forward)
        g = -r * 2 * np.imag(tensors.inner(f0.tensor, back.tensor) * np.conj(ol))

        grads.append(g)

    return grads


def state_angle_gradients(ket0: State, ket1: State, circ: Circuit) -> Sequence[float]:
    """
    Calculate the gradients of state angle for a parameterized quantum
    circuit, using the middle-out algorithm.

    Args:
        ket0: An initial state.
        ket1: A target state. We calculate the fidelity between this state and
            the resultant of the circuit.
        circ: A circuit that acts on ket0.
    Returns:
        The gradients of the inter-state angle with respect to the circuits
        parameters.
    """
    grads = state_fidelity_gradients(ket0, ket1, circ)
    fid = state_fidelity(circ.run(ket0), ket1)
    fid = np.real(fid)
    grads = -np.asarray(grads) / (2 * np.sqrt((1 - fid) * fid))
    return grads


def parameter_shift_circuits(
    circ: Circuit, index: int
) -> Tuple[float, Circuit, Circuit]:
    """
    Calculate the gradients of state angle for a parameterized quantum
    circuit, using the parameter-shift rule.

    Returns the gate shift-constant, and two circuits, circ0, circ1.
    Gradients are proportional to the difference in expectation between the two
    circuits.

    .. code-block:: python

        r, circ0, circ1 = qf.parameter_shift_circuits(circ, n)
        fid0 = qf.state_fidelity(circ0.run(ket0), ket1)
        fid1 = qf.state_fidelity(circ1.run(ket0), ket1)
        grad = r*(fid1-fid0)


    Args:
        circ: A quantum circuit with parameterized gates
        index: The index of the target gate in the quantum circuit
    Returns:
        r: The gate shift constant
        circ0: Circuit with parameter shift on target gate.
        circ1: Circuit with negative parameter shift on target gate.

    """

    elem = circ[index]
    assert isinstance(elem, Gate)
    gate_type = type(elem)
    if gate_type not in shift_constant:
        raise ValueError(_UNDIFFERENTIABLE_GATE_MSG)

    r = shift_constant[gate_type]
    param = list(elem.params)[0]
    gate0 = gate_type(param - 0.25 * pi / r, *elem.qubits)  # type: ignore
    circ0 = list(circ)
    circ0[index] = gate0

    gate1 = gate_type(param + 0.25 * pi / r, *elem.qubits)  # type: ignore
    circ1 = list(circ)
    circ1[index] = gate1

    return r, Circuit(circ0), Circuit(circ1)


# class SGD(object):
#     """Stochastic Gradient Descent optimizer"""
#     def __init__(self, learning_rate=0.001):
#         self.iterations = 0
#         self.learning_rate = learning_rate

#     def get_update(self, params, grads):
#         """ params and grads are list of numpy arrays
#         """
#         lr = self.learning_rate
#         new_params = []
#         for p, g in zip(params, grads):
#             new_params.append(p - lr * g)

#         self.iterations += 1
#         return new_params


# class Adam(object):
#     """Adam optimizer.

#     Args:
#         learning_rate: The learning rate
#         beta_1: The exponential decay rate for the 1st moment estimates.
#         beta_2: The exponential decay rate for the 2nd moment estimates.
#         epsilon: Numerical stability fuzz factor.

#     Returns:
#         bool: The return value. True for success, False otherwise.

#     Refs:
#         - [Adam - A Method for StochasticOptimization]
#             (http://arxiv.org/abs/1412.6980v8)
#     """

#     def __init__(
#         self,
#         learning_rate: float = 0.001,
#         beta_1: float = 0.9,
#         beta_2: float = 0.999,
#         epsilon: float = 1e-7,
#     ):

#         self.iterations = 0
#         self.learning_rate = learning_rate
#         self.beta_1 = beta_1
#         self.beta_2 = beta_2
#         self.epsilon = epsilon

#     def get_update(self, params: np.ndarray, grads: np.ndarray) -> Sequence[float]:
#         self.iterations += 1
#         t = self.iterations
#         lr = self.learning_rate
#         beta_1 = self.beta_1
#         beta_2 = self.beta_2
#         epsilon = self.epsilon

#         if not hasattr(self, "ms"):
#             self.ms = [np.zeros(p.shape) for p in params]
#             self.vs = [np.zeros(p.shape) for p in params]

#         lr_t = lr * np.sqrt(1.0 - beta_2 ** t) / (1.0 - beta_1 ** t)

#         new_params = []
#         for i in range(len(params)):
#             p = params[i]
#             g = grads[i]
#             m = self.ms[i]
#             v = self.vs[i]

#             m_t = (beta_1 * m) + (1.0 - beta_1) * g
#             v_t = (beta_2 * v) + (1.0 - beta_2) * g ** 2
#             p_t = p - lr_t * m_t / (np.sqrt(v_t) + epsilon)

#             self.ms[i] = m_t
#             self.vs[i] = v_t

#             new_params.append(p_t)

#         return new_params


# Fin



# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""Tests for quantumflow.xforest"""


import numpy as np
import pytest

import quantumflow as qf  # noqa: 402
from quantumflow import xforest  # noqa: 402

pytest.importorskip("pyquil")  # noqa: 402


def test_circuit_to_pyquil() -> None:
    circ = qf.Circuit()
    circ += qf.X(0)

    prog = xforest.circuit_to_pyquil(circ)
    assert str(prog) == "X 0\n"

    circ = qf.Circuit()
    circ1 = qf.Circuit()
    circ2 = qf.Circuit()
    circ1 += qf.Ry(np.pi / 2, 0)
    circ1 += qf.Rz(np.pi, 0)
    circ1 += qf.Ry(np.pi / 2, 1)
    circ1 += qf.Rx(np.pi, 1)
    circ1 += qf.CNot(0, 1)
    circ2 += qf.Rx(-np.pi / 2, 1)
    circ2 += qf.Ry(4.71572463191, 1)
    circ2 += qf.Rx(np.pi / 2, 1)
    circ2 += qf.CNot(0, 1)
    circ2 += qf.Rx(-2 * 2.74973750579, 0)
    circ2 += qf.Rx(-2 * 2.74973750579, 1)
    circ += circ1
    circ += circ2

    prog = xforest.circuit_to_pyquil(circ)
    new_circ = xforest.pyquil_to_circuit(prog)

    assert qf.circuits_close(circ, new_circ)


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import numpy as np
import pytest

from quantumflow import var

funcnames = ["arccos", "arcsin", "arctan", "cos", "exp", "sign", "sin", "sqrt", "tan"]


@pytest.mark.parametrize("funcname", funcnames)
def test_scalar_functions(funcname: str) -> None:
    x = var.Symbol("x")
    nx = 0.76  # Gives real answer for all functions
    subs = {"x": nx}

    symfn = getattr(var, funcname)
    npfn = getattr(np, funcname)

    assert np.isclose(var.asfloat(symfn(x), subs), npfn(nx))
    assert np.isclose(var.asfloat(symfn(nx), subs), npfn(nx))


def test_scalar_arctan2() -> None:
    x = var.Symbol("x")
    nx = 0.76
    y = var.Symbol("y")
    ny = -0.5
    subs = {"x": nx, "y": ny}

    assert np.isclose(var.asfloat(var.arctan2(x, y), subs), np.arctan2(nx, ny))
    assert np.isclose(var.asfloat(var.arctan2(nx, ny), subs), np.arctan2(nx, ny))


def test_almost_zero() -> None:
    assert var.almost_zero(0)
    assert var.almost_zero(0.0)
    assert var.almost_zero(0.000000000000000001)
    assert var.almost_zero(var.Symbol("x").evalf(subs={"x": 0}))


def test_isclose() -> None:
    assert var.isclose(1.0, 1.0000000001)
    assert not var.isclose(0.0, 0.0000000001, atol=0.000000000001)
    assert var.isclose(var.Symbol("x"), var.Symbol("x"))
    assert not var.isclose(var.Symbol("x"), 1.0)


def test_asexpression() -> None:
    s = var.asexpression(1.0)
    assert str(s) == "1"

    with pytest.raises(ValueError):
        s = var.asexpression(1.13434538345)

    s = var.asexpression(np.pi * 123)
    assert str(s) == "123*pi"

    s = var.asexpression(np.pi / 64)
    assert str(s) == "pi/64"

    s = var.asexpression(np.pi * 3 / 64)
    assert str(s) == "3*pi/64"

    s = var.asexpression(np.pi * 8 / 64)
    assert str(s) == "pi/8"

    s = var.asexpression(-np.pi * 3 / 8)
    assert str(s) == "-3*pi/8"

    s = var.asexpression(5 / 8)
    assert str(s) == "5/8"

    s = var.asexpression(2 / np.pi)
    assert str(s) == "2/pi"


# fin



#!/usr/bin/env python

# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QauntumFlow Examples
"""

import quantumflow as qf


def swap_test(ket, q0: qf.Qubit, q1: qf.Qubit, q2: qf.Qubit) -> qf.QubitTensor:
    """
    Apply a Swap Test to measure fidelity between qubits q1 and q2.

    Qubit q0 is an ancilla, which should be zero state initially. The qubits
    cannot be initially entangled.
    """
    circ = qf.Circuit()
    circ += qf.H(q0)
    circ += qf.CSwap(q0, q1, q2)
    circ += qf.H(q0)
    circ += qf.P0(q0)  # Measure
    ket = circ.run(ket)

    fid = 2 * ket.norm() - 1.0  # return fidelity
    return fid


if __name__ == "__main__":

    def main():
        """CLI"""
        print(swap_test.__doc__)

        print("Randomly generating two 1-qubit states...")

        ket0 = qf.zero_state([0])
        ket1 = qf.random_state([1])
        ket2 = qf.random_state([2])

        ket = qf.join_states(ket0, ket1)
        ket = qf.join_states(ket, ket2)

        fid = qf.state_fidelity(ket1, ket2.on(1))
        st_fid = swap_test(ket, 0, 1, 2)

        print("Fidelity:               ", fid)
        print("Fidelity from swap test:", st_fid)

    main()



# Copyright 2020-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import numpy as np
import pytest
from sympy import Symbol

import quantumflow as qf

# == test Gate ==


def test_gate_mul() -> None:
    # three cnots same as one swap
    gate0 = qf.IdentityGate([0, 1])

    gate1 = qf.CNot(1, 0)
    gate2 = qf.CNot(0, 1)
    gate3 = qf.CNot(1, 0)

    gate = gate1 @ gate0
    gate = gate2 @ gate
    gate = gate3 @ gate
    assert qf.gates_close(gate, qf.Swap(0, 1))

    # Again, but with labels
    gate0 = qf.IdentityGate(["a", "b"])

    gate1 = qf.CNot("b", "a")
    gate2 = qf.CNot("a", "b")
    gate3 = qf.CNot("b", "a")

    gate = gate1 @ gate0
    gate = gate2 @ gate
    gate = gate3 @ gate
    assert qf.gates_close(gate, qf.Swap("a", "b"))

    gate4 = qf.X("a")
    gate = gate4 @ gate

    with pytest.raises(NotImplementedError):
        gate = gate4 @ 3  # type: ignore


def test_gate_permute() -> None:
    gate0 = qf.CNot(0, 1)
    gate1 = qf.CNot(1, 0)

    assert not qf.gates_close(gate0, gate1)

    gate2 = gate1.permute([0, 1])
    assert gate2.qubits == (0, 1)
    assert qf.gates_close(gate1, gate2)

    gate3 = qf.ISwap(0, 1)
    gate4 = gate3.permute([1, 0])
    assert qf.gates_close(gate3, gate4)


def test_gates_evolve() -> None:
    rho0 = qf.zero_state(3).asdensity()
    qf.H(0).evolve(rho0)


def test_gate_H() -> None:
    gate0 = qf.X(0)
    assert gate0.cv_hermitian
    assert gate0.H is gate0

    gate1 = qf.ISwap(0, 1)
    assert not gate1.cv_hermitian
    assert qf.gates_close(gate1.H.H, gate1)


def test_su() -> None:
    su = qf.Swap(0, 1).su()
    assert np.isclose(np.linalg.det(su.asoperator()), 1.0)


def test_interchangeable() -> None:
    assert qf.Swap(0, 1).cv_interchangeable
    assert not qf.CNot(0, 1).cv_interchangeable


def test_gate_symbolic_params() -> None:
    theta = Symbol("θ")

    gate0 = qf.Rz(theta, 1)
    assert str(gate0) == "Rz(θ) 1"

    gate1 = gate0 ** 4
    assert str(gate1) == "Rz(4*θ) 1"

    circ = qf.Circuit([gate0, gate1])
    print(circ)
    diag = qf.circuit_to_diagram(circ)
    assert diag == "1: ───Rz(θ)───Rz(4*θ)───\n"

    gate2 = gate0.resolve({"θ": 2})
    assert gate2.param("theta") == 2.0

    with pytest.raises(KeyError):
        _ = gate2.param("asldfh")


def test_gate_rewire() -> None:
    gate0 = qf.CNot(1, 0)
    gate1 = gate0.on("B", "A")
    assert gate1.qubits == ("B", "A")

    gate2 = gate1.rewire({"A": "a", "B": "b", "C": "c"})
    assert gate2.qubits == ("b", "a")

    with pytest.raises(ValueError):
        gate1 = gate0.on("B", "A", "C")


def test_join_gates() -> None:
    gate = qf.join_gates(qf.H(0), qf.X(1))
    ket = qf.zero_state(2)
    ket = gate.run(ket)
    ket = qf.H(0).run(ket)
    ket = qf.X(1).run(ket)

    assert qf.states_close(ket, qf.zero_state(2))


# == test Unitary ==


def test_unitary_exceptions() -> None:
    tensor = qf.CNot(1, 0).tensor

    with pytest.raises(ValueError):
        qf.Unitary(tensor, [0])

    with pytest.raises(ValueError):
        qf.Unitary(tensor, [0, 1, 2])


# fin



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import atexit
import glob
import os
import re
import sys

import guzzle_sphinx_theme
from setuptools_scm import get_version

sys.path.insert(0, os.path.abspath(".."))

__version__ = get_version(root="..", relative_to=__file__)
print("__version__", __version__)

with open("./_templates/version.html", "w") as f:
    f.write('<div align="center">v%s</div>' % __version__)


# -- Project information -----------------------------------------------------

project = "QuantumFlow"
copyright = "2019, Gavin Crooks"
author = "Gavin Crooks"
release = __version__

html_title = "QuantumFlow Documentation"
html_short_title = "QuantumFlow"

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    "**": ["logo-text.html", "version.html", "searchbox.html", "globaltoc.html"]
}


def setup(app):
    app.add_stylesheet("qf.css")  # also can be a full URL


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx.ext.doctest",
    #  'sphinx_autodoc_typehints',
    "sphinx.ext.inheritance_diagram",
    "sphinxcontrib.bibtex",
]

extensions.append("guzzle_sphinx_theme")

napoleon_use_ivar = True
napoleon_use_rtype = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme_path = guzzle_sphinx_theme.html_theme_path()
html_theme = "guzzle_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


def do_edits():

    pats = [
        # Hacks to shorten type descriptors
        (r"quantumflow\.qubits", r"qf"),
        (r"quantumflow\.ops", r"qf"),
        (r"quantumflow\.stdops", r"qf"),
        (r"quantumflow\.gates\.gates_one", r"qf"),
        (r"quantumflow\.gates\.gates_two", r"qf"),
        (r"quantumflow\.gates\.gates_three", r"qf"),
        (r"quantumflow\.gates\.gates_qasm", r"qf"),
        (r"quantumflow\.gates", r"qf"),
        (r"quantumflow\.states", r"qf"),
        (r"quantumflow\.circuits", r"qf"),
        (r"quantumflow\.translate\.", r"qf."),
        (r"^/quantumflow\.", r"qf."),
        # (r'qf\.readthedocs\.io', r'quantumflow.readthedocs.io'),
        # Hacks to fixup types
        (r"sympy\.core\.symbol\.Symbol", r"Parameter"),
        (r"Sequence\[collections\.abc\.Hashable\]", "Qubits"),
        (r"collections\.abc\.Hashable", "Qubit"),
        (r"tensor: Any", r"tensor: TensorLike"),
        (r"&#x2192; Any", r"&#x2192; BKTensor"),
    ]

    files = glob.glob("*build/html/*.html")
    for filename in files:
        with open(filename, "r+") as f:
            text = f.read()
            for pattern, replacement in pats:
                text = re.sub(pattern, replacement, text)
            f.seek(0)
            f.truncate()
            f.write(text)

    print("Note: post sphinx text substitutions performed (conf.py)")


atexit.register(do_edits)



# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


"""
QuantumFlow representations of pure quantum states and actions on states.

.. contents:: :local:

State objects
#############
.. autoclass:: State
    :members:


Standard states
###############
.. autofunction:: zero_state
.. autofunction:: w_state
.. autofunction:: ghz_state
.. autofunction:: random_state


Actions on states
#################
.. autofunction:: join_states
.. autofunction:: print_state
.. autofunction:: print_probabilities

"""

from abc import ABC
from math import sqrt
from typing import Any, Dict, List, Mapping, TextIO, Tuple, TypeVar, Union

import numpy as np
import opt_einsum

from . import tensors, utils
from .qubits import Qubit, Qubits
from .tensors import QubitTensor, TensorLike

__all__ = [
    "State",
    "join_states",
    "print_probabilities",
    "print_state",
    "random_state",
    "ghz_state",
    "w_state",
    "zero_state",
    "Density",
    "mixed_density",
    "random_density",
    "join_densities",
]


QuantumStateType = TypeVar("QuantumStateType", bound="QuantumState")
"""TypeVar for quantum states"""


class QuantumState(ABC):
    def __init__(
        self, tensor: TensorLike, qubits: Qubits, memory: Mapping = None
    ) -> None:
        """
        Abstract base class for representations of a quantum state.

        Args:
            tensor: A vector or tensor of state amplitudes
            qubits: A sequence of qubit names.
                (Defaults to integer indices, e.g. [0, 1, 2] for 3 qubits)
            memory: Classical data storage. Stored as an immutable dictionary.
        """

        self._tensor = tensor
        self._qubits = tuple(qubits)

        if memory is None:
            self.memory: utils.FrozenDict = utils.FrozenDict()
        elif isinstance(memory, utils.FrozenDict):
            self.memory = memory
        else:
            self.memory = utils.FrozenDict(memory)

    @property
    def tensor(self) -> QubitTensor:
        """
        Returns the tensor representation of this state (if possible).
        """
        if self._tensor is None:
            raise ValueError("Cannot access quantum state")  # pragma: no cover
        return self._tensor

    @property
    def qubits(self) -> Qubits:
        """Return qubit labels of this state"""
        return self._qubits

    @property
    def qubit_nb(self) -> int:
        """Return the total number of qubits"""
        return len(self._qubits)

    def replace(
        self: QuantumStateType,
        *,
        tensor: TensorLike = None,
        qubits: Qubits = None,
        memory: Mapping = None,
    ) -> QuantumStateType:
        """
        Creates a copy of this state, replacing the fields specified.
        """
        # Interface similar to dataclasses.replace
        tensor = self.tensor if tensor is None else tensor
        qubits = self.qubits if qubits is None else qubits
        memory = self.memory if memory is None else memory
        return type(self)(tensor, qubits, memory)

    def store(self: QuantumStateType, *args: Any, **kwargs: Any) -> QuantumStateType:
        """Update information in classical memory and return a new State."""
        mem = self.memory.update(*args, **kwargs)
        return self.replace(memory=mem)

    # TESTME
    def on(self: QuantumStateType, *qubits: Qubit) -> QuantumStateType:
        """Return a copy of this State with new qubits"""
        return self.replace(qubits=qubits)

    # TESTME
    def rewire(self: QuantumStateType, labels: Dict[Qubit, Qubit]) -> QuantumStateType:
        """Relabel qubits and return copy of this Operation"""
        qubits = tuple(labels[q] for q in self.qubits)
        return self.on(*qubits)

    def permute(self: QuantumStateType, qubits: Qubits = None) -> QuantumStateType:
        """Return a copy of this state with state tensor transposed to
        put qubits in the given order. If an explicit qubit
        ordering isn't supplied, we put qubits in sorted order.
        """
        if qubits is None:
            qubits = sorted(self.qubits)
        tensor = tensors.permute(self.tensor, self.qubit_indices(qubits))
        return self.replace(tensor=tensor, qubits=qubits)

    def qubit_indices(self, qubits: Qubits) -> List[int]:
        """Convert qubits to index positions.

        Raises:
            ValueError: If argument qubits are not found in state qubits
        """
        return [self.qubits.index(q) for q in qubits]

    def norm(self) -> QubitTensor:
        """Return the state vector norm"""
        return tensors.norm(self.tensor)


class State(QuantumState):
    """The quantum state of a collection of qubits.

    Note that memory usage grows exponentially with the number of qubits.
    (16*2^N bytes for N qubits)

    """

    def __init__(
        self, tensor: TensorLike, qubits: Qubits = None, memory: Mapping = None
    ) -> None:
        """Create a new State from a tensor of qubit amplitudes

        Args:
            tensor: A vector or tensor of state amplitudes
            qubits: A sequence of qubit names.
                (Defaults to integer indices, e.g. [0, 1, 2] for 3 qubits)
            memory: Classical data storage. Stored as an immutable dictionary.
        """
        tensor = tensors.asqutensor(tensor)

        N = np.ndim(tensor)
        if qubits is None:
            qubits = range(N)
        elif len(qubits) != N:
            raise ValueError("Wrong number of qubits for tensor")

        super().__init__(tensor, qubits, memory)

    def normalize(self) -> "State":
        """Normalize the state"""
        tensor = self.tensor / np.sqrt(self.norm())
        return State(tensor, self.qubits, self.memory)

    def probabilities(self) -> QubitTensor:
        """
        Returns:
            The state probabilities
        """
        value = np.absolute(self.tensor)
        return value * value

    def sample(self, trials: int) -> np.ndarray:
        """Measure the state in the computational basis the the given number
        of trials, and return the counts of each output configuration.
        """
        # TODO: Can we do this within backend?
        probs = np.real(self.probabilities())
        res = np.random.multinomial(trials, probs.ravel())
        res = res.reshape(probs.shape)
        return res

    def expectation(
        self, diag_hermitian: TensorLike, trials: int = None
    ) -> QubitTensor:
        """Return the expectation of a measurement. Since we can only measure
        our computer in the computational basis, we only require the diagonal
        of the Hermitian in that basis.

        If the number of trials is specified, we sample the given number of
        times. Else we return the exact expectation (as if we'd performed an
        infinite number of trials. )
        """
        if trials is None:
            probs = self.probabilities()
        else:
            probs = np.real(tensors.asqutensor(self.sample(trials) / trials))

        diag_hermitian = tensors.asqutensor(diag_hermitian)
        return np.sum(np.real(diag_hermitian) * probs)

    def measure(self) -> TensorLike:
        """Measure the state in the computational basis.

        Returns:
            A [2]*bits array of qubit states, either 0 or 1
        """
        probs = np.real(self.probabilities())
        indices = np.asarray(list(np.ndindex(*[2] * self.qubit_nb)))
        res = np.random.choice(probs.size, p=probs.ravel())
        res = indices[res]
        return res

    def asdensity(self, qubits: Qubits = None) -> "Density":
        """Convert a pure state to a density matrix.

        Args:
            qubits: The qubit subspace. If not given return the density
                    matrix for all the qubits (which can take a lot of memory!)
        """
        N = self.qubit_nb

        if qubits is None:
            qubits = self.qubits

        contract_qubits: List[Qubit] = list(self.qubits)
        for q in qubits:
            contract_qubits.remove(q)

        left_subs = np.asarray(list(range(0, N)))
        right_subs = np.asarray(list(range(N, 2 * N)))

        indices = [self.qubits.index(qubit) for qubit in contract_qubits]
        for idx in indices:
            right_subs[idx] = left_subs[idx]

        left_tensor = self.tensor
        right_tensor = np.conj(left_tensor)

        tensor = opt_einsum.contract(left_tensor, left_subs, right_tensor, right_subs)

        return Density(tensor, qubits, self.memory)

    def __str__(self) -> str:
        state = self.tensor
        s = []
        count = 0
        MAX_ELEMENTS = 64
        for index, amplitude in np.ndenumerate(state):
            if not np.isclose(amplitude, 0.0):
                ket = "|" + "".join([str(n) for n in index]) + ">"
                s.append(f"({amplitude.real:0.04g}" f"{amplitude.imag:+0.04g}i) {ket}")
                count += 1
                if count > MAX_ELEMENTS:
                    s.append("...")
                    break
        return " + ".join(s)


# End class State


def zero_state(qubits: Union[int, Qubits]) -> State:
    """Return the all-zero state on N qubits"""
    N, qubits = _qubits_count_tuple(qubits)
    ket = np.zeros(shape=[2] * N)
    ket[(0,) * N] = 1
    return State(ket, qubits)


def w_state(qubits: Union[int, Qubits]) -> State:
    """Return a W state on N qubits"""
    N, qubits = _qubits_count_tuple(qubits)
    ket = np.zeros(shape=[2] * N)
    for n in range(N):
        idx = np.zeros(shape=N, dtype=int)
        idx[n] += 1
        ket[tuple(idx)] = 1 / sqrt(N)
    return State(ket, qubits)


def ghz_state(qubits: Union[int, Qubits]) -> State:
    """Return a GHZ state on N qubits"""
    N, qubits = _qubits_count_tuple(qubits)
    ket = np.zeros(shape=[2] * N)
    ket[(0,) * N] = 1 / sqrt(2)
    ket[(1,) * N] = 1 / sqrt(2)
    return State(ket, qubits)


def random_state(qubits: Union[int, Qubits]) -> State:
    """Return a random state from the space of N qubits"""
    N, qubits = _qubits_count_tuple(qubits)
    ket = np.random.normal(size=([2] * N)) + 1j * np.random.normal(size=([2] * N))
    return State(ket, qubits).normalize()


# == Actions on States ==


def join_states(ket0: State, ket1: State) -> State:
    """Join two mixed states into a larger qubit state"""
    qubits = tuple(ket0.qubits) + tuple(ket1.qubits)
    tensor = tensors.outer(ket0.tensor, ket1.tensor, rank=1)
    memory = ket0.memory.update(ket1.memory)  # TESTME
    return State(tensor, qubits, memory)


# = Output =

# TODO: clean up. Move to visualization?


def print_state(state: State, file: TextIO = None) -> None:
    """Print a state vector"""
    state = state.tensor
    for index, amplitude in np.ndenumerate(state):
        ket = "".join([str(n) for n in index])
        print(ket, ":", amplitude, file=file)


# TODO: Should work for density also. Check
def print_probabilities(state: State, ndigits: int = 4, file: TextIO = None) -> None:
    """
    Pretty print state probabilities.

    Args:
        state:
        ndigits: Number of digits of accuracy
        file: Output stream (Defaults to stdout)
    """
    prob = state.probabilities()
    for index, prob in np.ndenumerate(prob):
        prob = round(prob, ndigits)
        if prob == 0.0:
            continue
        ket = "".join([str(n) for n in index])
        print(ket, ":", prob, file=file)


# --  Mixed Quantum States --


class Density(QuantumState):
    """A density matrix representation of a mixed quantum state"""

    def __init__(
        self, tensor: TensorLike, qubits: Qubits = None, memory: Mapping = None
    ) -> None:
        tensor = tensors.asqutensor(tensor)

        N = np.ndim(tensor) // 2
        if qubits is None:
            qubits = range(N)
        elif len(qubits) != N:
            raise ValueError("Wrong number of qubits for tensor")

        super().__init__(tensor, qubits, memory)

    def trace(self) -> QubitTensor:
        """Return the trace of this density operator"""
        return tensors.trace(self.tensor, rank=2)

    def normalize(self) -> "Density":
        """Normalize state"""
        tensor = self.tensor / self.trace()
        return self.replace(tensor=tensor)

    # TESTME
    def probabilities(self) -> QubitTensor:
        """Returns: The state probabilities """
        prob = tensors.diag(self.tensor)
        return prob

    def asoperator(self) -> QubitTensor:
        """Return the density matrix as a square array"""
        return tensors.flatten(self.tensor, rank=2)

    def asdensity(self, qubits: Qubits = None) -> "Density":
        if qubits is None:
            return self
        tensor = tensors.partial_trace(self.tensor, self.qubit_indices(qubits))
        return Density(tensor, qubits, self.memory)


def mixed_density(qubits: Union[int, Qubits]) -> Density:
    """Returns the completely mixed density matrix"""
    N, qubits = _qubits_count_tuple(qubits)
    matrix = np.eye(2 ** N) / 2 ** N
    return Density(matrix, qubits)


def random_density(
    qubits: Union[int, Qubits], rank: int = None, ensemble: str = "Hilbert–Schmidt"
) -> Density:
    """
    Returns: A randomly sampled Density

    Args:
        qubits: A list or number of qubits.
        rank: Rank of density matrix. (Defaults to full rank)
        ensemble: Either 'Hilbert–Schmidt' (default) or 'Burr'

    Ref:
        - "Induced.info in the space of mixed quantum states" Karol
          Zyczkowski, Hans-Juergen Sommers, J. Phys. A34, 7111-7125 (2001)
          arXiv:quant-ph/0012101
        - "Random Bures mixed states and the distribution of their purity",
          Osipov, Sommers, and Zyczkowski, J. Phys. A: Math. Theor. 43,
          055302 (2010). arXiv:0909.5094
    """
    if ensemble == "Hilbert–Schmidt":
        return random_density_hs(qubits, rank)
    elif ensemble == "Bures":
        return random_density_bures(qubits, rank)
    raise ValueError(
        "Unknown ensemble. " "Valid Options are 'Hilbert–Schmidt' or 'Bures"
    )


# TODO: Check math
def random_density_hs(qubits: Union[int, Qubits], rank: int = None) -> Density:
    """
    Returns: A randomly sampled Density from the Hilbert–Schmidt
                ensemble of quantum states.

    Args:
        qubits: A list or number of qubits.
        rank: Rank of density matrix. (Defaults to full rank)

    Ref:
        "Induced.info in the space of mixed quantum states" Karol
        Zyczkowski, Hans-Juergen Sommers, J. Phys. A34, 7111-7125 (2001)
        arXiv:quant-ph/0012101
    """
    N, qubits = _qubits_count_tuple(qubits)
    size = (2 ** N, 2 ** N) if rank is None else (2 ** N, rank)

    X = utils.complex_ginibre_ensemble(size)
    matrix = X @ X.conj().T
    matrix /= np.trace(matrix)

    return Density(matrix, qubits=qubits)


# TODO: Check math
def random_density_bures(qubits: Union[int, Qubits], rank: int = None) -> Density:
    """
    Returns: A random Density drawn from the Bures measure.

    Args:
        qubits: A list or number of qubits.
        rank: Rank of density matrix. (Defaults to full rank)

    Ref:
        "Random Bures mixed states and the distribution of their purity",
         Osipov, Sommers, and Zyczkowski, J. Phys. A: Math. Theor. 43,
         055302 (2010). arXiv:0909.5094
    """
    N, qubits = _qubits_count_tuple(qubits)
    dim = 2 ** N
    size = (dim, dim) if rank is None else (dim, rank)
    P = np.eye(dim) + utils.unitary_ensemble(dim)
    G = utils.complex_ginibre_ensemble(size=size)
    B = P @ G @ G.conj().T @ P.conj().T
    B /= np.trace(B)

    return Density(B, qubits=qubits)


def join_densities(rho0: Density, rho1: Density) -> Density:
    """Join two mixed states into a larger qubit state"""
    qubits = tuple(rho0.qubits) + tuple(rho1.qubits)
    tensor = tensors.outer(rho0.tensor, rho1.tensor, rank=2)
    memory = rho0.memory.update(rho1.memory)  # TESTME
    return Density(tensor, qubits, memory)


def _qubits_count_tuple(qubits: Union[int, Qubits]) -> Tuple[int, Qubits]:
    """Utility method for unraveling 'qubits: Union[int, Qubits]' arguments"""
    if isinstance(qubits, int):
        return qubits, tuple(range(qubits))
    return len(qubits), qubits


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Unit tests for quantumflow States
"""


import io

import numpy as np
import pytest

import quantumflow as qf
from quantumflow.utils import FrozenDict

from .config_test import REPS


def test_zeros() -> None:
    ket = qf.zero_state(4)
    assert ket.tensor[0, 0, 0, 0] == 1
    assert ket.tensor[0, 0, 1, 0] == 0


def test_w_state() -> None:
    ket = qf.w_state(4)
    assert ket.tensor[0, 0, 0, 0] == 0
    assert ket.tensor[0, 0, 1, 0] * 2.0 == 1


def test_ghz_state() -> None:
    ket = qf.ghz_state(4)
    assert ket.tensor[0, 0, 0, 1] == 0
    assert np.isclose(2.0 * ket.tensor[0, 0, 0, 0] ** 2.0, 1.0)


def test_random_state() -> None:
    state = qf.random_state(4)
    assert state.tensor.shape == (2,) * 4


def test_state_bits() -> None:
    for n in range(1, 6):
        assert qf.zero_state(n).qubit_nb == n


def test_state_labels() -> None:
    # Quil labeling convention
    N = 4
    qubits = range(N - 1, -1, -1)
    ket = qf.zero_state(qubits)
    ket = qf.X(0).run(ket)
    ket = qf.X(1).run(ket)
    assert ket.tensor[0, 0, 1, 1] == 1

    ket = ket.on(0, 1, 3, 4)
    assert ket.tensor[0, 0, 1, 1] == 1

    ket = ket.permute([4, 3, 0, 1])

    ket = ket.permute()
    assert ket.qubits == (0, 1, 3, 4)

    ket = ket.rewire({0: 0, 1: 1, 3: 3, 4: 5})
    assert ket.qubits == (0, 1, 3, 5)


def test_probability() -> None:
    state = qf.w_state(3)
    qf.print_state(state)
    prob = state.probabilities()

    qf.print_probabilities(state)
    assert np.isclose(prob.sum(), 1)


def test_states_close() -> None:
    ket0 = qf.w_state(4)
    ket1 = qf.w_state(3)
    ket2 = qf.w_state(4)
    ket3 = qf.ghz_state(3)

    assert qf.states_close(ket0, ket2)
    assert not qf.states_close(ket1, ket3)
    assert qf.states_close(ket2, ket2)


def test_str() -> None:
    ket = qf.random_state(10)
    s = str(ket)
    assert s[-3:] == "..."


def test_print_state() -> None:
    f = io.StringIO()
    state = qf.w_state(5)
    qf.print_state(state, file=f)


def test_print_probabilities() -> None:
    f = io.StringIO()
    state = qf.w_state(5)
    qf.print_probabilities(state, file=f)


def test_measure() -> None:
    ket = qf.zero_state(2)
    res = ket.measure()
    assert np.allclose(res, [0, 0])

    ket = qf.X(0).run(ket)
    res = ket.measure()
    assert np.allclose(res, [1, 0])

    ket = qf.H(0).run(ket)
    ket = qf.CNot(0, 1).run(ket)
    for _ in range(REPS):
        res = ket.measure()
        assert res[0] == res[1]  # Both qubits measured in same state


def test_sample() -> None:
    ket = qf.zero_state(2)
    ket = qf.H(0).run(ket)
    ket = qf.CNot(0, 1).run(ket)

    samples = ket.sample(10)
    assert samples.sum() == 10


def test_expectation() -> None:
    ket = qf.zero_state(1)
    ket = qf.H(0).run(ket)

    m = ket.expectation([0.4, 0.6])
    assert np.isclose(m, 0.5)

    m = ket.expectation([0.4, 0.6], 10)


def test_random_density() -> None:
    rho = qf.random_density(4)
    assert list(rho.tensor.shape) == [2] * 8

    rho1 = qf.random_density(4, rank=2, ensemble="Hilbert–Schmidt")
    assert list(rho1.tensor.shape) == [2] * 8

    rho2 = qf.random_density(4, rank=2, ensemble="Bures")
    assert list(rho2.tensor.shape) == [2] * 8

    with pytest.raises(ValueError):
        qf.random_density(4, rank=2, ensemble="not_an_ensemble")


def test_density() -> None:
    ket = qf.random_state(3)
    matrix = np.outer(ket.tensor, np.conj(ket.tensor))
    qf.Density(matrix)
    rho = qf.Density(matrix, [0, 1, 2])

    with pytest.raises(ValueError):
        qf.Density(matrix, [0, 1, 2, 3])

    assert rho.asdensity() is rho

    rho = rho.on(10, 11, 12).permute([12, 11, 10])
    assert rho.qubits == (12, 11, 10)

    rho.permute()


def test_state_to_density() -> None:
    density = qf.ghz_state(4).asdensity()
    assert density.tensor.shape == (2,) * 8

    prob = density.probabilities()
    assert np.isclose(prob[0, 0, 0, 0], 0.5)
    assert np.isclose(prob[0, 1, 0, 0], 0)
    assert np.isclose(prob[1, 1, 1, 1], 0.5)

    ket = qf.random_state(3)
    density = ket.asdensity()
    ket_prob = ket.probabilities()
    density_prob = density.probabilities()

    for index, prob in np.ndenumerate(ket_prob):
        assert np.isclose(prob, density_prob[index])


def test_density_trace() -> None:
    rho = qf.random_density(3)
    assert np.isclose(rho.trace(), 1)

    rho = qf.Density(np.eye(8))
    assert np.isclose(rho.trace(), 8)

    rho = rho.normalize()
    assert np.isclose(rho.trace(), 1)


def test_mixed_density() -> None:
    rho = qf.mixed_density(4)
    assert rho.trace() == 1


def test_join_densities() -> None:
    rho0 = qf.zero_state([0]).asdensity()
    rho1 = qf.zero_state([1]).asdensity()
    rho01 = qf.join_densities(rho0, rho1)
    assert rho01.qubits == (0, 1)


def test_memory() -> None:
    ket0 = qf.zero_state(1)
    assert ket0.memory == {}

    ro = ["ro[0]", "ro[1]"]
    ket1 = ket0.store({ro[1]: 1})
    assert ket0.memory == {}
    assert ket1.memory == {ro[1]: 1}

    ket2 = qf.H(0).run(ket1)
    assert ket2.memory == ket1.memory

    ket3 = ket2.store({ro[1]: 0, ro[0]: 0})
    assert ket3.memory == {ro[0]: 0, ro[1]: 0}

    N = 4
    wf = np.zeros(shape=[2] * N)
    wf[(0,) * N] = 1
    ket = qf.State(wf, list(range(N)), {"a": 2})
    assert ket.memory == {"a": 2}
    assert isinstance(ket.memory, FrozenDict)


def test_density_memory() -> None:
    rho0 = qf.zero_state(1).asdensity()
    assert rho0.memory == {}

    ro = ["ro[0]", "ro[1]"]
    rho1 = rho0.store({ro[1]: 1})
    assert rho0.memory == {}
    assert rho1.memory == {ro[1]: 1}

    rho2 = qf.H(0).aschannel().evolve(rho1)
    assert rho2.memory == rho2.memory

    rho3 = qf.X(0).evolve(rho1)
    assert rho3.memory == rho2.memory


def test_join_states() -> None:
    q0 = qf.zero_state([0])
    q1 = qf.zero_state([1, 2])
    q2 = qf.join_states(q0, q1)
    assert q2.qubit_nb == 3
    assert q2.tensor[0, 0, 0] == 1

    q3 = qf.zero_state([3, 4, 5])
    q3 = qf.X(4).run(q3)

    q4 = qf.join_states(qf.join_states(q0, q1), q3)
    assert q4.qubit_nb == 6
    assert q4.tensor[0, 0, 0, 0, 1, 0] == 1


def test_normalize() -> None:
    ket = qf.random_state(2)
    assert np.isclose(ket.norm(), 1.0)

    ket = qf.P0(0).run(ket)
    assert ket.norm() < 1.0
    ket = ket.normalize()
    assert np.isclose(ket.norm(), 1.0)


def test_expectation_again() -> None:
    ket = qf.zero_state(4)
    M = np.zeros(shape=([2] * 4))
    M[0, 0, 0, 0] = 42
    M[1, 0, 0, 0] = 1
    M[0, 1, 0, 0] = 2
    M[0, 0, 1, 0] = 3
    M[0, 0, 0, 1] = 4

    avg = ket.expectation(M)
    assert avg == 42

    ket = qf.w_state(4)
    assert ket.expectation(M) == 2.5


def test_error() -> None:
    ket = qf.zero_state(4)
    with pytest.raises(ValueError):
        qf.State(ket.tensor, [0, 1, 2])


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


# Standard Gates

# FIXME: Ordering of gates in docs

"""
.. contents:: :local:
.. currentmodule:: quantumflow


Standard gates
##############

.. autoclass:: StdGate
    :members:


Standard one-qubit gates
************************
.. autoclass:: I
.. autoclass:: X
.. autoclass:: Y
.. autoclass:: Z
.. autoclass:: H
.. autoclass:: S
.. autoclass:: T
.. autoclass:: PhaseShift
.. autoclass:: RX
.. autoclass:: RY
.. autoclass:: RZ
.. autoclass:: Ph

Standard two-qubit gates
************************
.. autoclass:: CZ
.. autoclass:: CNOT
.. autoclass:: SWAP
.. autoclass:: ISWAP
.. autoclass:: CPHASE00
.. autoclass:: CPHASE01
.. autoclass:: CPHASE10
.. autoclass:: CPHASE
.. autoclass:: PSWAP


Standard three-qubit gates
**************************
.. autoclass:: CCNot
.. autoclass:: CSwap



Additional gates
################


One-qubit gates
***************
.. autoclass:: S_H
.. autoclass:: T_H
.. autoclass:: XPow
.. autoclass:: YPow
.. autoclass:: ZPow
.. autoclass:: HPow
.. autoclass:: V
.. autoclass:: V_H
.. autoclass:: PhasedX
.. autoclass:: PhasedXPow


Two-qubit gates
***************
.. autoclass:: CAN
.. autoclass:: XX
.. autoclass:: YY
.. autoclass:: ZZ
.. autoclass:: XY
.. autoclass:: EXCH
.. autoclass:: Barenco
.. autoclass:: CY
.. autoclass:: CH
.. autoclass:: CNotPow
.. autoclass:: SqrtISwap
.. autoclass:: SqrtISwap_H
.. autoclass:: SqrtSwap
.. autoclass:: SqrtSwap_H


Three-qubit gates
*****************
.. autoclass:: CZ
.. autoclass:: Deutsch
.. autoclass:: CCiX


QASM gates
**********
.. autoclass::  U3
.. autoclass::  U2
.. autoclass::  CU3
.. autoclass::  CRZ
.. autoclass::  RZZ


Mapping between APIs
####################
Each of the main quantum computing python APIs (QuantumFlow, Cirq, qsikit
(QASM), and pyQuil) have different gates available and different naming
conventions. The following table maps gate names between these APIs.

==========================  =========== =============== =========== =========== ===========
Description                 QF          Cirq            QASM/qsikit PyQuil      Pennylane
==========================  =========== =============== =========== =========== ===========
Identity  (single qubit)    I           I               id or iden  I
Pauli-X                     X           X               x           X           PauliX
Pauli-Y                     Y           Y               y           Y           PauliY
Pauli-Z                     Z           Z               z           Z           PauliZ
Hadamard                    H           H               h           H           Hadamard
X-rotations                 Rx          rx              rx          RX          RX
Y-rotations                 Ry          ry              ry          RY          RY
Z-rotations                 Rz          rz              rz          RZ          RZ
Sqrt of Z                   S           S               s           S           S
Sqrt of S                   T           T               t           T           T
Phase shift                 PhaseShift  .               u1          PHASE       PhaseShift
Bloch rotations             Rn          .               .           .           .
Powers of X                 XPow        XPowGate        .           .           .
Powers of Y                 YPow        YPowGate        .           .           .
Powers of Z                 ZPow        ZPowGate        .           .           .
Powers of Hadamard          HPow        HPowGate        .           .           .
Inv. of S                   S_H         .               sdg         .           .
Inv. of T                   T_H         .               tdg         .           .
Sqrt of X                   V           .               .           .           .
Inv. sqrt of X              V_H         .               .           .           .

Powers of X⊗X               XX          XXPowGate       .           .           .
Powers of Y⊗Y               YY          YYPowGate       .           .           .
Powers of Z⊗Z               ZZ          ZZPowGate       .           .           .
Canonical                   Can         .               .           .           .
Controlled-Not              CNOT        CNOT            cx          CNOT        CNOT
Controlled-Z                CZ          CZ              cz          CZ          CZ
Controlled-Y                CY          .               cy          .           .
Controlled-Hadamard         CH          .               ch          .           .
Controlled-V                CV          .               .           .           .
Controlled-inv-V            CV_H        .               .           .           .
Powers of CNOT				CXPow		CNotPowGate		.			.			.
Powers of CY				CYPow			.				.			.			.
Powers of CZ				CZPow		CZPowGate		.			.			.
Swap                        Swap        SWAP            swap        .           SWAP
Exchange                    Exch        SwapPowGate (*) .           .
iSwap                       ISwap       ISWAP           .           ISWAP       .
XY (powers of iSwap)        XY          ISwapPowGate(*) .           XY(*)       .
Givens rotation				Givens		givens			.			.			.
Barenco                     Barenco     .               .           .           .
B (Berkeley)                B           .               .           .           .
Sqrt-iSWAP                  SqrtISwap   .               .           .           .
Inv. of sqrt-iSWAP          SqrtISwap_H .               .           .           .
Sqrt-SWAP                   SqrtSwap    .               .           .           .
Inv. of sqrt-SWAP           SqrtSwap_H  .               .           .           .
ECP                         ECP         .               .           .           .
W (Dual-rail Hadamard)      W           .               .           .           .

Toffoli                     CCNot       CCX             ccx         CCNOT       Toffoli
Fredkin                     CSwap       CSWAP           cswap       CSWAP       CSWAP
Controlled-Controlled-Z     CCZ         CCZ             .           .           .
Deutsch                     Deutsch     .               .           .           .
Powers of CCNOT             CCXPow    	CCXPowGate      .           .           .
Powers of CCZ               .           CCZPowGate      .           .           .

* Forest specific gates
Controlled-phase            CPhase      CZPowGate(*)    cu1         CPHASE      .
Controlled-phase on 00      CPhase00    .               .           CPHASE00    .
Controlled-phase on 01      CPhase01    .               .           CPHASE01    .
Controlled-phase on 10      CPhase10    .               .           CPHASE10    .
Parametric Swap             PSwap       .               .           PSWAP       .

* Cirq specific gates
Fermionic-Simulation        FSim        FSimGate        .           .           .
Phased-X gate               PhasedX     .               .           .           .
Powers of Phased-X gate     PhasedXPow  PhasedXPowGate  .           .           .
Sycamore                    Sycamore    Sycamore        .           .           .
Fermionic swap              FSwap       FSwap           .           .           .
Powers of fermionic swap    FSwapPow    FSwapPow        .           .           .

* QASM/qiskit gates
QASM's U3 gate              U3          .               u3          .           Rot
QASM's U2 gate              U2          .               u2          .
QASM's controlled-U3        CU3         .               cu3         .           .
QASM's ZZ-rotations         RZZ         .               rzz         .           CRot
Controlled-RX               .           .               .           .           CRX
Controlled-RY               .           .               .           .           CRY
Controlled-RZ               CRZ         .               crz                     CRZ
==========================  =========== =============== =========== =========== ===========

(*) Modulo differences in parametrization

(**) Cirq defines XX, YY, and ZZ gates, as XX^1, YY^1, ZZ^1 which are direct
products of single qubit gates.
(e.g. XX(t,0,1) is the same as X(0) X(1) when t=1)


"""  # noqa: E501
# Note: see comments in gates_qasm.py to understand the QASM gate mappings.


from .stdgates_1q import *  # noqa: F401, F403
from .stdgates_2q import *  # noqa: F401, F403
from .stdgates_3q import *  # noqa: F401, F403
from .stdgates_cirq import *  # noqa: F401, F403
from .stdgates_forest import *  # noqa: F401, F403
from .stdgates_qasm import *  # noqa: F401, F403



# Copyright 2019-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import random

import numpy as np

import quantumflow as qf

from ..config_test import REPS


def test_U3() -> None:
    theta = 0.2
    phi = 2.3
    lam = 1.1

    gate3 = qf.Circuit(
        [qf.U3(theta, phi, lam, 0), qf.U3(theta, phi, lam, 0).H]
    ).asgate()
    assert qf.almost_identity(gate3)

    gate2 = qf.Circuit([qf.U2(phi, lam, 0), qf.U2(phi, lam, 0).H]).asgate()
    assert qf.almost_identity(gate2)


def test_cu1() -> None:
    # Test that QASM's cu1 gate is the same as CPHASE up to global phase

    for _ in range(REPS):
        theta = random.uniform(0, 4)
        circ0 = qf.Circuit(
            [
                qf.Rz(theta / 2, 0),
                qf.CNot(0, 1),
                qf.Rz(-theta / 2, 1),
                qf.CNot(0, 1),
                qf.Rz(theta / 2, 1),
            ]
        )
        gate0 = circ0.asgate()
        gate1 = qf.CPhase(theta, 0, 1)
        assert qf.gates_close(gate0, gate1)


def test_CU3() -> None:
    theta = 0.2
    phi = 2.3
    lam = 1.1

    gate = qf.Circuit(
        [qf.CU3(theta, phi, lam, 0, 1), qf.CU3(theta, phi, lam, 0, 1).H]
    ).asgate()
    assert qf.almost_identity(gate)

    cgate = qf.ControlGate([0], qf.U3(theta, phi, lam, 1))
    assert qf.gates_close(qf.CU3(theta, phi, lam, 0, 1), cgate)


def test_CRZ() -> None:
    theta = 0.23
    gate0 = qf.CRZ(theta, 0, 1)
    coords = qf.canonical_coords(gate0)
    assert np.isclose(coords[0], 0.5 * theta / np.pi)
    assert np.isclose(coords[1], 0.0)
    assert np.isclose(coords[2], 0.0)

    coords = qf.canonical_coords(gate0 ** 3.3)
    assert np.isclose(coords[0], 3.3 * 0.5 * theta / np.pi)

    gate1 = qf.Circuit([qf.CRZ(theta, 0, 1), qf.CRZ(theta, 0, 1).H]).asgate()
    assert qf.almost_identity(gate1)


def test_RZZ() -> None:
    theta = 0.23
    gate0 = qf.RZZ(theta, 0, 1)
    gate1 = qf.ZZ(theta / np.pi, 0, 1)
    assert qf.gates_close(gate0, gate1)
    assert qf.gates_close(gate0.H, gate1.H)
    assert qf.gates_close(gate0 ** 0.12, gate1 ** 0.12)


# fin



# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
.. contents:: :local:
.. currentmodule:: quantumflow


Circuit objects
###############
.. autoclass:: Circuit
    :members:

.. autoclass:: DAGCircuit
    :members:


Standard circuits
#################


.. autofunction:: control_circuit
.. autofunction:: zyz_circuit
.. autofunction:: euler_circuit
.. autofunction:: phase_estimation_circuit
.. autofunction:: addition_circuit
.. autofunction:: ghz_circuit
.. autofunction:: graph_circuit
.. autofunction:: graph_circuit_params

Visualizations
##############

.. autofunction:: circuit_to_diagram
.. autofunction:: circuit_to_image
.. autofunction:: circuit_to_latex
.. autofunction:: latex_to_image
"""

import textwrap
from collections import defaultdict
from itertools import chain
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    overload,
)

import networkx as nx
import numpy as np

from .config import CIRCUIT_INDENT
from .ops import Channel, Gate, Operation
from .qubits import Qubit, Qubits
from .states import Density, State, zero_state
from .stdgates import CZ, ZZ, CCNot, CNot, H, X, XPow, YPow, ZPow
from .var import Variable

__all__ = [
    "Circuit",
    "count_operations",
    "map_gate",
    "control_circuit",
    "euler_circuit",
    "zyz_circuit",
    "phase_estimation_circuit",
    "addition_circuit",
    "ghz_circuit",
    "graph_circuit",
    "graph_circuit_params",
    "graph_state_circuit",
]


class Circuit(Sequence, Operation):
    """A quantum Circuit contains a sequences of circuit elements.
    These can be any quantum Operation, including other circuits.
    """

    def __init__(self, elements: Iterable[Operation] = ()) -> None:
        elements = tuple(elements)
        qbs = [q for elem in elements for q in elem.qubits]  # gather
        qbs = list(set(qbs))  # unique
        qbs = sorted(qbs)  # sort

        super().__init__(qubits=qbs)
        self._elements: Tuple[Operation, ...] = elements

    # Methods for Sequence
    @overload
    def __getitem__(self, key: int) -> Operation:
        ...  # pragma: no cover

    @overload  # noqa: F811
    def __getitem__(self, key: slice) -> "Circuit":
        ...  # pragma: no cover

    def __getitem__(self, key: Union[int, slice]) -> Operation:  # noqa: F811s
        if isinstance(key, slice):
            return Circuit(self._elements[key])
        return self._elements[key]

    def __len__(self) -> int:
        return len(self._elements)

    def add(self, other: Iterable[Operation]) -> "Circuit":
        """Concatenate operations and return new circuit"""
        return Circuit(chain(self, other))

    def __add__(self, other: "Circuit") -> "Circuit":
        return self.add(other)

    def __iadd__(self, other: Iterable[Any]) -> "Circuit":
        return self.add(other)

    def __iter__(self) -> Iterator[Operation]:
        yield from self._elements

    # End methods for Sequence

    def flat(self) -> Iterator[Operation]:
        """Iterate over all elementary elements of Circuit,
        recursively flattening composite elements such as
        sub-Circuit's, DAGCircuit's, and Moment's"""
        for elem in self:
            if hasattr(elem, "flat"):
                yield from elem.flat()  # type: ignore
            else:
                yield from elem

    def size(self) -> int:
        """Return the number of operations in this circuit"""
        return len(self._elements)

    def on(self, *qubits: Qubit) -> "Circuit":
        if len(qubits) != self.qubit_nb:
            raise ValueError("Wrong number of qubits")
        labels = dict(zip(self.qubits, qubits))

        return self.rewire(labels)

    def rewire(self, labels: Dict[Qubit, Qubit]) -> "Circuit":
        return Circuit([elem.rewire(labels) for elem in self])

    def run(self, ket: State = None) -> State:
        """
        Apply the action of this circuit upon a state.

        If no initial state provided an initial zero state will be created.
        """
        if ket is None:
            qubits = self.qubits
            ket = zero_state(qubits=qubits)

        for elem in self:
            ket = elem.run(ket)
        return ket

    def evolve(self, rho: Density = None) -> Density:
        if rho is None:
            qubits = self.qubits
            rho = zero_state(qubits=qubits).asdensity()

        for elem in self:
            rho = elem.evolve(rho)
        return rho

    def asgate(self) -> Gate:
        from .modules import IdentityGate

        gate: Gate = IdentityGate(self.qubits)
        for elem in self:
            gate = elem.asgate() @ gate
        return gate

    def aschannel(self) -> Channel:
        from .modules import IdentityGate

        chan = IdentityGate(self.qubits).aschannel()
        for elem in self:
            chan = elem.aschannel() @ chan
        return chan

    @property
    def H(self) -> "Circuit":
        """Returns the Hermitian conjugate of this circuit.
        If all the subsidiary gates are unitary, returns the circuit inverse.
        """
        return Circuit([elem.H for elem in reversed(self)])

    def __str__(self) -> str:
        circ_str = "\n".join([str(elem) for elem in self])
        circ_str = textwrap.indent(circ_str, " " * CIRCUIT_INDENT)
        return "\n".join([self.name, circ_str])

    # TESTME
    def resolve(self, subs: Mapping[str, float]) -> "Circuit":
        """Resolve the parameters of all of the elements of this circuit"""
        return Circuit(op.resolve(subs) for op in self)

    @property
    def params(self) -> Tuple[Variable, ...]:
        return tuple(item for elem in self for item in elem.params)

    def param(self, name: str) -> Variable:
        raise ValueError("Cannot lookup parameters by name for composite operations")

    # TESTME
    def specialize(self) -> "Circuit":
        """Specialize all of the elements of this circuit"""
        return Circuit([elem.specialize() for elem in self])

    def _repr_png_(self) -> Optional[bytes]:  # pragma: no cover
        """Jupyter/IPython rich display"""
        from .visualization import circuit_to_image

        try:
            return circuit_to_image(self)._repr_png_()
        except (ValueError, IOError):
            return None

    def _repr_html_(self) -> Optional[str]:  # pragma: no cover
        """Jupyter/IPython rich display"""
        from .visualization import circuit_to_diagram

        diag = circuit_to_diagram(self)
        return '<pre style="line-height: 90%">' + diag + "</pre>"


# End class Circuit


def count_operations(elements: Iterable[Operation]) -> Dict[Type[Operation], int]:
    """Return a count of different operation types given a collection of
    operations, such as a Circuit or DAGCircuit
    """
    op_count: Dict[Type[Operation], int] = defaultdict(int)
    for elem in elements:
        op_count[type(elem)] += 1
    return dict(op_count)


# TODO: MapGate? # TODO: Rename map_gate_circuit
def map_gate(gate: Gate, args: Sequence[Qubits]) -> Circuit:
    """Applies the same gate to all input qubits in the argument list.

    >>> circ = qf.map_gate(qf.H(), [[0], [1], [2]])
    >>> print(circ)
    Circuit
        H(0)
        H(1)
        H(2)

    """
    circ = Circuit()

    for qubits in args:
        circ += gate.on(*qubits)

    return circ


def control_circuit(controls: Qubits, gate: Gate) -> Circuit:
    """
    Returns a circuit for a target gate controlled by
    a collection of control qubits. [Barenco1995]_

    Uses a number of gates quadratic in the number of control qubits.

    .. [Barenco1995] A. Barenco, C. Bennett, R. Cleve (1995) Elementary Gates
        for Quantum Computation`<https://arxiv.org/abs/quant-ph/9503016>`_
        Sec 7.2
    """
    # Kudos: Adapted from Rigetti Grove's utility_programs.py
    # grove/utils/utility_programs.py::ControlledProgramBuilder

    from .modules import ControlGate

    circ = Circuit()
    if len(controls) == 1:
        q0 = controls[0]
        if isinstance(gate, X):
            circ += CNot(q0, gate.qubits[0])
        else:
            cgate = ControlGate([q0], gate)
            circ += cgate
    else:
        circ += control_circuit(controls[-1:], gate ** 0.5)
        circ += control_circuit(controls[0:-1], X(controls[-1]))
        circ += control_circuit(controls[-1:], gate ** -0.5)
        circ += control_circuit(controls[0:-1], X(controls[-1]))
        circ += control_circuit(controls[0:-1], gate ** 0.5)
    return circ


def zyz_circuit(t0: Variable, t1: Variable, t2: Variable, q0: Qubit) -> Circuit:
    """A circuit of ZPow, XPow, ZPow gates on the same qubit"""
    return euler_circuit(t0, t1, t2, q0, "ZYZ")


def euler_circuit(
    t0: Variable,
    t1: Variable,
    t2: Variable,
    q0: Qubit,
    euler: str = "ZYZ",
) -> Circuit:
    """
    A Euler decomposition of a 1-qubit gate.

    The 'euler' argument can be used to specify any of the 6 Euler
    decompositions: 'XYX', 'XZX', 'YXY', 'YZY', 'ZXZ', 'ZYZ' (Default)
    """
    euler_circ = {
        "XYX": (XPow, YPow, XPow),
        "XZX": (XPow, ZPow, XPow),
        "YXY": (YPow, XPow, YPow),
        "YZY": (YPow, ZPow, YPow),
        "ZXZ": (ZPow, XPow, ZPow),
        "ZYZ": (ZPow, YPow, ZPow),
    }

    gate0, gate1, gate2 = euler_circ[euler]
    circ = Circuit()
    circ += gate0(t0, q0)
    circ += gate1(t1, q0)
    circ += gate2(t2, q0)
    return circ


def phase_estimation_circuit(gate: Gate, outputs: Qubits) -> Circuit:
    """Returns a circuit for quantum phase estimation.

    The gate has an eigenvector with eigenvalue e^(i 2 pi phase). To
    run the circuit, the eigenvector should be set on the gate qubits,
    and the output qubits should be in the zero state. After evolution and
    measurement, the output qubits will be (approximately) a binary fraction
    representation of the phase.

    The output registers can be converted with the aid of the
    quantumflow.utils.bitlist_to_int() method.

    >>> import numpy as np
    >>> import quantumflow as qf
    >>> N = 8
    >>> phase = 1/4
    >>> gate = qf.Rz(-4*np.pi*phase, N)
    >>> circ = qf.phase_estimation_circuit(gate, range(N))
    >>> res = circ.run().measure()[0:N]
    >>> est_phase = int(''.join([str(d) for d in res]), 2) / 2**N # To float
    >>> print(phase, est_phase)
    0.25 0.25

    """
    from .modules import ControlGate

    circ = Circuit()
    circ += map_gate(H(0), list(zip(outputs)))  # Hadamard on all output qubits

    for cq in reversed(outputs):
        cgate = ControlGate([cq], gate)
        circ += cgate
        gate = gate @ gate

    from .modules import InvQFTGate

    circ += InvQFTGate(outputs).decompose()

    return circ


def addition_circuit(addend0: Qubits, addend1: Qubits, carry: Qubits) -> Circuit:
    """Returns a quantum circuit for ripple-carry addition. [Cuccaro2004]_

    Requires two carry qubit (input and output). The result is returned in
    addend1.

    .. [Cuccaro2004]
        A new quantum ripple-carry addition circuit, Steven A. Cuccaro,
        Thomas G. Draper, Samuel A. Kutin, David Petrie Moulton
        arXiv:quant-ph/0410184 (2004)
    """

    if len(addend0) != len(addend1):
        raise ValueError("Number of addend qubits must be equal")

    if len(carry) != 2:
        raise ValueError("Expected 2 carry qubits")

    def _maj(qubits: Qubits) -> Circuit:
        q0, q1, q2 = qubits
        circ = Circuit()
        circ += CNot(q2, q1)
        circ += CNot(q2, q0)
        circ += CCNot(q0, q1, q2)
        return circ

    def _uma(qubits: Qubits) -> Circuit:
        q0, q1, q2 = qubits
        circ = Circuit()
        circ += CCNot(q0, q1, q2)
        circ += CNot(q2, q0)
        circ += CNot(q0, q1)
        return circ

    qubits = (
        [carry[0]]
        + list(chain.from_iterable(zip(reversed(addend1), reversed(addend0))))
        + [carry[1]]
    )

    circ = Circuit()

    for n in range(0, len(qubits) - 3, 2):
        circ += _maj(qubits[n : n + 3])

    circ += CNot(qubits[-2], qubits[-1])

    for n in reversed(range(0, len(qubits) - 3, 2)):
        circ += _uma(qubits[n : n + 3])

    return circ


def ghz_circuit(qubits: Qubits) -> Circuit:
    """Returns a circuit that prepares a multi-qubit Bell state from the zero
    state.
    """
    circ = Circuit()

    circ += H(qubits[0])
    for q0 in range(0, len(qubits) - 1):
        circ += CNot(qubits[q0], qubits[q0 + 1])

    return circ


def graph_circuit_params(
    topology: nx.Graph, steps: int, init_bias: float = 0.0, init_scale: float = 0.01
) -> Sequence[float]:
    """Return a set of initial parameters for graph_circuit()"""
    N = len(topology.nodes())
    K = len(topology.edges())
    total = N + N * 2 * (steps + 1) + K * steps
    params = np.random.normal(loc=init_bias, scale=init_scale, size=[total])

    return params


def graph_circuit(
    topology: nx.Graph,
    steps: int,
    params: Sequence[float],
) -> Circuit:
    """
    Create a multilayer parameterized circuit given a graph of connected
    qubits.

    We alternate between applying a sublayer of arbitrary 1-qubit gates to all
    qubits, and a sublayer of ZZ gates between all connected qubits.

    In practice the pattern is ZPow, XPow, ZPow, (ZZ, XPow, ZPow )*, where ZPow are
    1-qubit Z rotations, and XPow and 1-qubits X rotations. Since a Z rotation
    commutes across a ZZ, we only need one Z sublayer per layer.

    Our fundamental 2-qubit interaction is the Ising like ZZ gate. We could
    apply a more general gate, such as the universal Canonical gate. But ZZ
    gates commute with each other, whereas other choice of gate would not,
    which would necessitate specifying the order of all 2-qubit gates within
    the layer.
    """

    def tx_layer(topology: nx.Graph, layer_params: Sequence[float]) -> Circuit:
        circ = Circuit()
        for p, q0 in zip(layer_params, topology.nodes()):
            circ += XPow(p, q0)
        return circ

    def tz_layer(topology: nx.Graph, layer_params: Sequence[float]) -> Circuit:
        circ = Circuit()
        for p, q0 in zip(layer_params, topology.nodes()):
            circ += ZPow(p, q0)
        return circ

    def zz_layer(topology: nx.Graph, layer_params: Sequence[float]) -> Circuit:
        circ = Circuit()
        for p, (q0, q1) in zip(layer_params, topology.edges()):
            circ += ZZ(p, q0, q1)
        return circ

    N = len(topology.nodes())
    K = len(topology.edges())

    circ = Circuit()

    n = 0
    circ += tz_layer(topology, params[n : n + N])
    n += N
    circ += tx_layer(topology, params[n : n + N])
    n += N
    circ += tz_layer(topology, params[n : n + N])
    n += N

    for _ in range(steps):
        circ += zz_layer(topology, params[n : n + K])
        n += K
        circ += tx_layer(topology, params[n : n + N])
        n += N
        circ += tz_layer(topology, params[n : n + N])
        n += N

    return circ


def graph_state_circuit(topology: nx.Graph) -> Circuit:
    """
    Return a circuit to create a graph state, given a
    particular graph topology.

    Refs:
        - [Wikipedia: Graph State](https://en.wikipedia.org/wiki/Graph_state)
    """
    circ = Circuit()

    for q in topology.nodes:
        circ += H(q)

    for q0, q1 in topology.edges:
        circ += CZ(q0, q1)

    return circ


# Fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow: Gates specific to QASM
"""

import numpy as np

from .. import tensors
from ..config import CTRL
from ..ops import StdGate
from ..paulialgebra import Pauli, sZ
from ..qubits import Qubit
from ..tensors import QubitTensor
from ..utils import cached_property
from ..var import Variable
from .stdgates_1q import PhaseShift, Rz
from .stdgates_2q import CNot
from .stdgates_forest import CPhase

__all__ = ("U3", "U2", "CU3", "CRZ", "RZZ")


# Is QASM's U1 gate a PhaseShift gate or an RZ gate?
# This is very confusing. In the QASM paper U1(lam) is defined as both
# PhaseShift(lam) (Eq. 3) and as U3(0,0, lam), which is RZ(lam) (Bottom of
# page 10). https://arxiv.org/pdf/1707.03429.pdf
# The same happens in the code.
# https://github.com/Qiskit/qiskit-terra/blob/master/qiskit/extensions/standard/u1.py
# U1._define() assumes U1 is RZ, but U1._matrix() gives the PhaseShift gate.
# In code and paper, RZ is defined as U1, so it's even not clear if QASM's RZ
# is our PhaseShift or RZ!
# Phase shift and RZ only differ by a global phase, so this doesn't matter most
# of the time. An exception is constructing controlled gates. Looking at QASM's
# control gates, crz is in fact a controlled-RZ, and cu1 is a
# controlled-PhaseShift! (up to global phase).
# Therefore, QASM's u1 is a PhaseShift gate, and rz is an RZ gate,
# cu1 is a CPHASE.
# u3 and cu3 are even more of a nightmare. See notes below.

U1 = PhaseShift
CU1 = CPhase


class U3(StdGate):
    r"""The U3 single qubit gate from QASM.
    The U2 gate is the U3 gate with theta=pi/2. The U1 gate has theta=phi=0,
    which is the same as a PhaseShift gate.

    ..math::
        \text{U3}(\theta, \phi, \lambda) = R_z(\phi) R_y(\theta) R_z(\lambda)

    Refs:
        https://arxiv.org/pdf/1707.03429.pdf (Eq. 2)
        https://github.com/Qiskit/qiskit-terra/blob/master/qiskit/extensions/standard/u3.py
    """

    def __init__(
        self, theta: Variable, phi: Variable, lam: Variable, q0: Qubit
    ) -> None:
        super().__init__(params=[theta, phi, lam], qubits=[q0])

    @cached_property
    def tensor(self) -> QubitTensor:
        theta = self.float_param("theta")
        phi = self.float_param("phi")
        lam = self.float_param("lam")

        unitary = [
            [np.cos(theta / 2.0), -np.sin(theta / 2.0) * np.exp(1j * lam)],
            [
                np.sin(theta / 2.0) * np.exp(1j * phi),
                np.cos(theta / 2.0) * np.exp(1j * (phi + lam)),
            ],
        ]
        return tensors.asqutensor(unitary)

        # Alternative definition from paper
        # Differs by a phase from definition above, but this matches
        # definition of CU3 in qsikit (but not definition of CU3 in paper.)
        # circ = Circuit()
        # circ += RZ(lam)
        # circ += RY(theta)
        # circ += RZ(phi)
        # return circ.asgate().tensor

    @property
    def H(self) -> "U3":
        theta, phi, lam = self.params
        return U3(-theta, -lam, -phi, *self.qubits)


# end class U3


class U2(StdGate):
    """A 'single pulse' 1-qubit gate defined in QASM"""

    def __init__(self, phi: Variable, lam: Variable, q0: Qubit) -> None:
        super().__init__(params=[phi, lam], qubits=[q0])

    @cached_property
    def tensor(self) -> QubitTensor:
        phi, lam = self.params
        return U3(np.pi / 2, phi, lam, q0=0).tensor

    @property
    def H(self) -> "U3":
        phi, lam = self.params
        return U3(-np.pi / 2, -lam, -phi, *self.qubits)


# end class U2


class CU3(StdGate):
    r"""The controlled U3 gate, as defined by QASM.

    Ref:
        https://github.com/Qiskit/qiskit-terra/blob/master/qiskit/extensions/standard/cu3.py
    """
    _diagram_labels = [CTRL, "U3({theta}, {phi}, {lam})"]

    def __init__(
        self,
        theta: Variable,
        phi: Variable,
        lam: Variable,
        q0: Qubit,
        q1: Qubit,
    ) -> None:
        super().__init__(params=[theta, phi, lam], qubits=[q0, q1])

    @cached_property
    def tensor(self) -> QubitTensor:
        theta = self.float_param("theta")
        phi = self.float_param("phi")
        lam = self.float_param("lam")

        # Note: Gate is defined via this circuit in QASM
        # Except for first line, which was added to qsikit to make the
        # definitions of cu3 and u3 in qsikit consistent.
        # https://github.com/Qiskit/qiskit-terra/pull/2755
        # That seems silly. They should have fixed the phase of u3 to match
        # the definition in the QASM paper, not change the cu3 gate to
        # something entirely different.
        from ..circuits import Circuit

        circ = Circuit(
            [
                PhaseShift((lam + phi) / 2, 0),
                PhaseShift((lam - phi) / 2, 1),
                CNot(0, 1),
                U3(-theta / 2, 0.0, -(phi + lam) / 2, 1),
                CNot(0, 1),
                U3(theta / 2, phi, 0.0, 1),
            ]
        )

        return circ.asgate().tensor

    @property
    def H(self) -> "CU3":
        theta, phi, lam = self.params
        return CU3(-theta, -lam, -phi, *self.qubits)


# end class class CU3


class CRZ(StdGate):
    r"""A controlled RZ gate."""
    cv_tensor_structure = "diagonal"
    _diagram_labels = [CTRL, "Rz({theta})"]

    def __init__(self, theta: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[theta], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        (theta,) = self.params
        q0, q1 = self.qubits
        return Rz(theta, q1).hamiltonian * (1 - sZ(q0)) / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        (theta,) = self.params
        q0, q1 = self.qubits
        gate = Rz(theta, q1)
        from ..modules import ControlGate

        return ControlGate([q0], gate).tensor

    @property
    def H(self) -> "CRZ":
        return self ** -1

    def __pow__(self, t: Variable) -> "CRZ":
        theta = self.param("theta")
        return CRZ(theta * t, *self.qubits)


# end class CRZ


# TODO: Check proper phase, so can add Hamiltonian.
# How is this different from CRZ?
class RZZ(StdGate):
    """A two-qubit ZZ-rotation gate, as defined by QASM.
    Same as ZZ(theta/pi), up to phase.
    """

    cv_interchangeable = True
    cv_tensor_structure = "diagonal"

    def __init__(self, theta: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[theta], qubits=[q0, q1])

    @cached_property
    def tensor(self) -> QubitTensor:
        (theta,) = self.params
        q0, q1 = self.qubits
        from ..circuits import Circuit

        circ = Circuit([CNot(q0, q1), PhaseShift(theta, q1), CNot(q0, q1)])
        return circ.asgate().tensor

    @property
    def H(self) -> "RZZ":
        return self ** -1

    def __pow__(self, e: Variable) -> "RZZ":
        (theta,) = self.params
        return RZZ(theta * e, *self.qubits)


# end class RZZ


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import quantumflow as qf


def test_compile() -> None:
    circ0 = qf.addition_circuit([0], [1], [2, 3])
    circ1 = qf.compile_circuit(circ0)
    assert qf.circuits_close(circ0, circ1)
    assert circ1.size() == 76

    dagc = qf.DAGCircuit(circ1)
    assert dagc.depth(local=False) == 16
    counts = qf.count_operations(dagc)
    assert counts[qf.ZPow] == 27
    assert counts[qf.XPow] == 32
    assert counts[qf.CZ] == 17


def test_merge() -> None:
    circ0 = qf.Circuit(
        [
            qf.XPow(0.4, 0),
            qf.XPow(0.2, 0),
            qf.YPow(0.1, 1),
            qf.YPow(0.1, 1),
            qf.ZPow(0.1, 1),
            qf.ZPow(0.1, 1),
        ]
    )
    dagc = qf.DAGCircuit(circ0)

    qf.merge_tx(dagc)
    qf.merge_tz(dagc)
    qf.merge_ty(dagc)
    circ1 = qf.Circuit(dagc)
    assert len(circ1) == 3


# fin



#!/usr/bin/env python

# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


"""
QuantumFlow Examples:
    Demonstration of decomposing a CSwap into CZ and 1-qubit gates
"""
import quantumflow as qf

circ0 = qf.Circuit([qf.CSwap(0, 1, 2)])

translators = [
    qf.translate_cswap_to_ccnot,
    qf.translate_ccnot_to_cnot,
    qf.translate_cnot_to_cz,
]
circ1 = qf.circuit_translate(circ0, translators)

assert circ1.size() == 33
print(qf.circuit_diagram(circ1))



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


"""
.. contents:: :local:
.. currentmodule:: quantumflow.xqiskit

Interface between IBM's Qiskit and QuantumFlow

.. autoclass:: QiskitSimulator
.. autofunction:: qiskit_to_circuit
.. autofunction:: circuit_to_qiskit
"""

import qiskit

from .circuits import Circuit
from .ops import Operation, StdGate
from .qubits import Qubits
from .states import State
from .stdops import If, Initialize
from .translate import circuit_translate, select_translators
from .utils import invert_map

# This module imports qiskit, so we do not include it at top level.
# Must be imported explicitly. e.g.
# > from quantumflow.xqiskit import qiskit_to_circuit, circuit_to_qiskit
#
# Note that QASM specific gates are defined in quantumflow/stdgates/stdgates_qasm.py
# Conceivable you might want to use those gates in QuantumFlow without loading
# qiskit

__all__ = [
    "QiskitSimulator",
    "qiskit_to_circuit",
    "circuit_to_qiskit",
    "translate_gates_to_qiskit",
]

QASM_TO_QF = {
    "ccx": "CCNot",
    "ch": "CH",
    "crz": "CRZ",
    "cswap": "CSwap",
    "cu1": "CPhase",
    "cu3": "CU3",
    "cx": "CNot",
    "cy": "CY",
    "cz": "CZ",
    "h": "H",
    "id": "I",
    "rx": "Rx",
    "ry": "Ry",
    "rz": "Rz",
    "rzz": "RZZ",
    "s": "S",
    "sdg": "S_H",
    "swap": "Swap",
    "t": "T",
    "tdg": "T_H",
    "u1": "PhaseShift",
    "u2": "U2",
    "u3": "U3",
    "x": "X",
    "y": "Y",
    "z": "Z",
    # 'barrier': 'Barrier',   # TODO TESTME
    # 'measure': 'Measure'   # TODO
    #  'initialize': 'Initialize',
}
"""Map from qiskit operation names to QuantumFlow names"""


NAMED_GATES = StdGate.cv_stdgates

# TODO: 'multiplexer', 'snapshot', 'unitary'


class QiskitSimulator(Operation):
    def __init__(self, *elements: Operation) -> None:
        self._circuit = Circuit(*elements)

    @property
    def qubits(self) -> Qubits:
        return self._circuit.qubits

    def run(self, ket: State = None) -> State:
        circ = self._circuit
        if ket is not None:
            circ = Circuit(Initialize(ket)) + circ

        qkcircuit = circuit_to_qiskit(circ, translate=True)
        simulator = qiskit.Aer.get_backend("statevector_simulator")
        job = qiskit.execute(qkcircuit, simulator)
        result = job.result()
        tensor = result.get_statevector()

        res = State(tensor, list(reversed(self.qubits)))

        if ket is not None:
            res = res.permute(ket.qubits)
        return res


def qiskit_to_circuit(qkcircuit: qiskit.QuantumCircuit) -> Circuit:
    """Convert a qsikit QuantumCircuit to QuantumFlow's Circuit"""
    # We assume that there is only one quantum register of qubits.

    named_ops = dict(StdGate.cv_stdgates)

    circ = Circuit()

    for instruction, qargs, cargs in qkcircuit:
        name = instruction.name
        if name not in QASM_TO_QF:
            raise NotImplementedError("Unknown qiskit operation")

        qf_name = QASM_TO_QF[name]
        qubits = [q.index for q in qargs]
        args = [float(param) for param in instruction.params] + qubits
        gate = named_ops[qf_name](*args)  # type: ignore

        if instruction.condition is None:
            circ += gate
        else:
            classical, value = instruction.condition
            circ += If(gate, classical, value)

    return circ


# TODO: Default translate to false?
# TODO: QISKIT_GATES
def circuit_to_qiskit(circ: Circuit, translate: bool = True) -> qiskit.QuantumCircuit:
    """Convert a QuantumFlow's Circuit to a qsikit QuantumCircuit."""

    # In qiskit each gate is defined as a class, and then a method is
    # monkey patched onto QuantumCircuit which will create that gate and
    # append it to the circuit. The method names correspond to the qasm
    # names in QASM_TO_QF

    if translate:
        circ = translate_gates_to_qiskit(circ)

    QF_TO_QASM = invert_map(QASM_TO_QF)
    # QF_TO_QASM["I"] = "i"

    # We assume only one QuantumRegister. Represent qubits by index in register
    qreg = qiskit.QuantumRegister(circ.qubit_nb)
    qubit_map = {q: qreg[i] for i, q in enumerate(circ.qubits)}

    qkcircuit = qiskit.QuantumCircuit(qreg)

    for op in circ:
        if op.name == "Initialize":
            name = "initialize"

            # TODO: CHECK
            params = op.tensor.transpose().flatten()

            qbs = [qubit_map[qb] for qb in op.qubits]
            getattr(qkcircuit, name)(params, qbs)
        else:
            name = QF_TO_QASM[op.name]
            params = [float(p) for p in op.params]
            qbs = [qubit_map[qb] for qb in op.qubits]
            getattr(qkcircuit, name)(*params, *qbs)

        # TODO: Handle If, Reset, ...

    return qkcircuit


def translate_gates_to_qiskit(circ: Circuit) -> Circuit:
    """Convert QF gates to gates understood by qiskit"""
    target_gates = list([NAMED_GATES[n] for n in QASM_TO_QF.values()])
    trans = select_translators(target_gates)  # type: ignore

    circ = circuit_translate(circ, trans)
    return circ


# fin



#!/usr/bin/env python

# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""Canonical decompositions of pairs of 2-qubit gates."""

import argparse
from sys import platform as sys_pf

from matplotlib import pyplot, rcParams
from mpl_toolkits.mplot3d import Axes3D

if sys_pf == "darwin":
    import matplotlib

    matplotlib.use("TkAgg")

import quantumflow as qf  # noqa: E402

__version__ = qf.__version__
__description__ = """Canonical decompositions of pairs of 2-qubit gates.

We construct 2-qubit gate sandwiches, where two random 1-qubit gates
are sandwiched between two 2-qubit gates, which are specified by their
canonical coordinates). We then decompose to find the composite gate's
canonical coordinates within the Weyl chamber. A 3D view of the data
is constructed and shown with matplotlib.

See arXiv:1904.10541 "Fixed-Depth Two-Qubit Circuits and the Monodromy
Polytope" by Eric C. Peterson, Gavin E. Crooks, and Robert S. Smith.
For an introduction to the Weyl Chamber of 2-qubit gates, see
http://threeplusone.com/weyl and http://threeplusone.com/gates

Canonical Coordinates of 2-qubit Gates
--------------------------------------
The intrinsic non-local character of a 2-qubit gate can be captured by
the gate canonical coordinates with the Weyl chamber.

0    0    0   Identity   [Also (1, 0, 0)]
.5   0    0   CNOT / CZ
.5   .5   0   ISWAP / DCNOT
.5   .5   .5  SWAP

.25  0    0   sqrt(CNOT)  [Also (.75, 0, 0)]
.25  .25  0   sqrt(ISWAP) [Also (.75, .25, 0)]
.25  .25  .25 sqrt(SWAP)
.75  .25  .25 Adjoint of sqrt(SWAP)

.5   .25  0   B
.5   .25  .25 ECP
.375 .375 0 DB (Dagwood Bumstead)
.5   .5   .25 QFT (Parametric swap, half way between ISWAP and SWAP)


Interesting Sandwiches
----------------------

* CZ-CZ sandwich
Special orthongal gates. Bottom of Weyl chamber.
    ./weyl.py 0.5 0 0   0.5 0 0

* CZ-ISWAP sandwich
Improper orthongal gates. Plane connecting CZ, ISWAP, and SWAP.
    ./weyl.py 0.5 0 0   0.5 0.5 0

* sqrt(ISWAP) sandwich
Tetrahedron with the ECP gate as top point.
    ./weyl.py 0.25 0.25 0   0.25 0.25 0

* Dagwood Bumstead (DB) sandwich
Of all gates in the XY family, the Dagwood Bumstead gate makes the
biggest sandwich. (3/4 of all 2-qubit gates)
    ./weyl.py .375 .375 0  .375 .375 0

"""
__author__ = "Gavin E. Crooks"


SAMPLES = 1024 * 2


def sandwich_decompositions(coords0, coords1, samples=SAMPLES):
    """Create composite gates, decompose, and return a list
    of canonical coordinates"""
    decomps = []
    for _ in range(samples):
        circ = qf.Circuit()
        circ += qf.Can(*coords0, 0, 1)
        circ += qf.RandomGate([0])
        circ += qf.RandomGate([1])
        circ += qf.Can(*coords1, 0, 1)
        gate = circ.asgate()

        coords = qf.canonical_coords(gate)
        decomps.append(coords)

    return decomps


def display_weyl(decomps):
    """Construct and display 3D plot of canonical coordinates"""
    tx, ty, tz = list(zip(*decomps))

    rcParams["axes.labelsize"] = 24
    rcParams["font.family"] = "serif"
    rcParams["font.serif"] = ["Computer Modern Roman"]
    rcParams["text.usetex"] = True
    fig = pyplot.figure()
    ax = Axes3D(fig)
    ax.scatter(tx, ty, tz)
    ax.plot((1,), (1,), (1,))
    ax.plot(
        (0, 1, 1 / 2, 0, 1 / 2, 1, 1 / 2, 1 / 2),
        (0, 0, 1 / 2, 0, 1 / 2, 0, 1 / 2, 1 / 2),
        (0, 0, 0, 0, 1 / 2, 0, 0, 1 / 2),
    )
    ax.plot(
        (0, 1 / 2, 1, 1 / 2, 1 / 2),
        (0, 1 / 4, 0, 1 / 4, 1 / 2),
        (0, 1 / 4, 0, 1 / 4, 0),
    )

    points = [
        (0, 0, 0),
        (1 / 4, 0, 0),
        (1 / 2, 0, 0),
        (3 / 4, 0, 0),
        (1, 0, 0),
        (1 / 4, 1 / 4, 0),
        (1 / 2, 1 / 4, 0),
        (3 / 4, 1 / 4, 0),
        (1 / 2, 1 / 2, 0),
        (1 / 4, 1 / 4, 1 / 4),
        (1 / 2, 1 / 4, 1 / 4),
        (3 / 4, 1 / 4, 1 / 4),
        (1 / 2, 1 / 2, 1 / 4),
        (1 / 2, 1 / 2, 1 / 2),
    ]

    ax.scatter(*zip(*points))
    eps = 0.04
    ax.text(0, 0, 0 - 2 * eps, "I", ha="center")
    ax.text(1, 0, 0 - 2 * eps, "I", ha="center")
    ax.text(1 / 2, 1 / 2, 0 - 2 * eps, "iSWAP", ha="center")
    ax.text(1 / 2, 1 / 2, 1 / 2 + eps, "SWAP", ha="center")
    ax.text(1 / 2, 0, 0 - 2 * eps, "CNOT", ha="center")

    # More coordinate labels
    # ax.text(1/4-eps, 1/4, 1/4, '$\sqrt{SWAP}$', ha='right')
    # ax.text(3/4+eps, 1/4, 1/4, '$\sqrt{SWAP}^\dagger$', ha='left')
    # ax.text(1/4, 0, 0-2*eps, '$\sqrt{{CNOT}}$', ha='center')
    # ax.text(3/4, 0, 0-2*eps, '$\sqrt{{CNOT}}$', ha='center')
    # ax.text(1/2, 1/4, 0-2*eps, 'B', ha='center')
    # ax.text(1/2, 1/4, 1/4+eps, 'ECP', ha='center')
    # ax.text(1/4, 1/4, 0-2*eps, '$\sqrt{iSWAP}$', ha='center')
    # ax.text(3/4, 1/4, 0-2*eps, '$\sqrt{iSWAP}$', ha='center')
    # ax.text(1/2, 1/2+eps, 1/4, 'PSWAP(1/2)', ha='left')

    ax.set_xlim(0, 1)
    ax.set_ylim(-1 / 4, 3 / 4)
    ax.set_zlim(-1 / 4, 3 / 4)

    # Get rid of the panes
    ax.w_xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    ax.w_yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    ax.w_zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))

    # Get rid of the spines
    ax.w_xaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
    ax.w_yaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
    ax.w_zaxis.line.set_color((1.0, 1.0, 1.0, 0.0))

    # Get rid of the ticks
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks([])

    pyplot.show()


# ---------- Command Line Interface ----------
def _cli():

    parser = argparse.ArgumentParser(
        description=__description__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "coords",
        nargs=6,
        metavar="COORD",
        type=float,
        action="store",
        help="Coordinates of initial and final gates",
    )

    # Parse
    opts = vars(parser.parse_args())
    coords = opts.pop("coords")

    # Decompose and display
    decomps = sandwich_decompositions(coords[0:3], coords[3:6])
    display_weyl(decomps)


if __name__ == "__main__":
    _cli()



#!/usr/bin/env python

import random
import time
import timeit

import quantumflow as qf
from quantumflow import xcirq

GATES = 512
REPS = 8
QUBITS = 16


def benchmark_circuit(N, gate_nb, gate):
    """Create a random circuit with N qubits and given number of gates.
    Half the gates will be 1-qubit, and half 2-qubit
    """
    qubits = list(range(0, N))
    circ = qf.Circuit()
    K = gate.qubit_nb

    for _ in range(0, gate_nb):
        qbs = random.sample(qubits, K)
        circ += gate.on(*qbs)

    return circ


def _cli():

    gates = [
        qf.I(0),
        qf.X(0),
        qf.Y(0),
        qf.Z(0),
        qf.S(0),
        qf.T(0),
        qf.H(0),
        qf.XPow(0.2, 0),
        qf.YPow(0.2, 0),
        qf.ZPow(0.2, 0),
        qf.CNot(0, 1),
        qf.CZ(0, 1),
        qf.Swap(0, 1),
        qf.ISwap(0, 1),
        qf.CCNot(0, 1, 2),
        qf.CCZ(0, 1, 2),
        qf.CSwap(0, 1, 2),
    ]

    print()
    print("Gate     QF GOPS         Cirq GOPS")

    # for n in range(4):
    #     circ = benchmark_circuit(QUBITS, GATES, qf.RandomGate([0,1]))
    #     t = timeit.timeit(lambda: circ.run(), number=REPS,
    #                       timer=time.process_time)
    #     gops = int((GATES*REPS)/t)
    #     gops = int((gops * 100) + 0.5) / 100.0
    #     print(f"gate qubits: {n}  gops:{gops}")

    for gate in gates:
        circ = benchmark_circuit(QUBITS, GATES, gate)
        t = timeit.timeit(lambda: circ.run(), number=REPS, timer=time.process_time)

        cq = xcirq.CirqSimulator(circ)
        t2 = timeit.timeit(lambda: cq.run(), number=REPS, timer=time.process_time)

        gops = int((GATES * REPS) / t)
        gops = int((gops * 100) + 0.5) / 100.0

        gops2 = int((GATES * REPS) / t2)
        gops2 = int((gops2 * 100) + 0.5) / 100.0

        if gops / gops2 > 0.8:
            print(gate.name, "\t", gops, "\t", gops2)
        else:
            print(gate.name, "\t", gops, "\t", gops2, "\t☹️")


if __name__ == "__main__":
    _cli()



# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
.. module:: quantumflow.xforest

Interface to pyQuil and the Rigetti Forest.

.. autofunction:: circuit_to_pyquil
.. autofunction:: pyquil_to_circuit


.. autodata:: QUIL_TO_QF

"""

from itertools import chain

from pyquil.quil import Program as pqProgram
from pyquil.quilbase import Declare as pqDeclare
from pyquil.quilbase import Gate as pqGate
from pyquil.quilbase import Halt as pqHalt
from pyquil.quilbase import Measurement as pqMeasurement
from pyquil.quilbase import Pragma as pqPragma

from . import utils
from .circuits import Circuit
from .ops import Gate, StdGate
from .stdops import Measure

__all__ = [
    "QUIL_TO_QF",
    "circuit_to_pyquil",
    "pyquil_to_circuit",
]


QUIL_TO_QF = {
    "I": "I",
    "X": "X",
    "Y": "Y",
    "Z": "Z",
    "H": "H",
    "S": "S",
    "T": "T",
    "PhaseShift": "PhaseShift",
    "RX": "Rx",
    "RY": "Ry",
    "RZ": "Rz",
    "CZ": "CZ",
    "CNOT": "CNot",
    "SWAP": "Swap",
    "ISWAP": "ISwap",
    "CPHASE00": "CPhase00",
    "CPHASE01": "CPhase01",
    "CHPHASE10": "CPhase10",
    "CPHASE": "CPhase",
    "PSWAP": "PSwap",
    "CCNOT": "CCNot",
    "CSWAP": "CSwap",
}
"""Map from QUIL operation names to QuantumFlow names"""


def circuit_to_pyquil(circuit: Circuit) -> pqProgram:
    """Convert a QuantumFlow circuit to a pyQuil program"""
    prog = pqProgram()

    QF_TO_QUIL = utils.invert_map(QUIL_TO_QF)
    for elem in circuit:
        if isinstance(elem, Gate) and elem.name in QF_TO_QUIL:
            params = list(elem.params)
            name = QF_TO_QUIL[elem.name]
            prog.gate(name, params, elem.qubits)
        else:
            raise ValueError("Cannot convert operation to pyquil")  # pragma: no cover

    return prog


def pyquil_to_circuit(program: pqProgram) -> Circuit:
    """Convert a protoquil pyQuil program to a QuantumFlow Circuit"""

    circ = Circuit()
    for inst in program.instructions:
        # print(type(inst))
        if isinstance(inst, pqDeclare):  # Ignore
            continue  # pragma: no cover
        if isinstance(inst, pqHalt):  # Ignore
            continue  # pragma: no cover
        if isinstance(inst, pqPragma):  # Ignore
            continue  # pragma: no cover
        elif isinstance(inst, pqMeasurement):  # pragma: no cover
            circ += Measure(inst.qubit.index)
        elif isinstance(inst, pqGate):
            name = QUIL_TO_QF[inst.name]
            defgate = StdGate.cv_stdgates[name]
            qubits = [q.index for q in inst.qubits]
            gate = defgate(*chain(inst.params, qubits))  # type: ignore
            circ += gate
        else:
            raise ValueError("PyQuil program is not protoquil")  # pragma: no cover

    return circ


# fin
